﻿'    SmartDongle ClientForm.vb
'
'    MicroWorks, Inc.
'    2808 North Cole Road
'    Boise, ID 83704
'
'    www.smartdongle.com
'
'    ***** BEGIN LICENSE BLOCK *****
'
'    Version: MPL 1.1/GPL 2.0/LGPL 2.1
'
'    The contents of this file are subject to the Mozilla Public License
'    Version 1.1 (the "License"); you may not use this file except in
'    compliance with the License. You may obtain a copy of the License at
'    http://www.mozilla.org/MPL/
'
'    Software distributed under the License is distributed on an "AS IS" basis,
'    WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
'    for the specific language governing rights and limitations under the
'    License.
'
'    The Original Code is MicroWorks, Inc. code.
'
'    The Initial Developer of the Original Code is MicroWorks, Inc. Portions
'    created by MicroWorks, Inc. are Copyright (C) 2010
'    MicroWorks, Inc. All Rights Reserved.
'
'    Contributor(s):
'
'    Alternatively, the contents of this file may be used under the terms of
'    either the GNU General Public License Version 2 or later (the "GPL"), or
'    the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
'    in which case the provisions of the GPL or the LGPL are applicable instead
'    of those above. If you wish to allow use of your version of this file only
'    under the terms of either the GPL or the LGPL, and not to allow others to
'    use your version of this file under the terms of the MPL, indicate your
'    decision by deleting the provisions above and replace them with the notice
'    and other provisions required by the GPL or the LGPL. If you do not delete
'    the provisions above, a recipient may use your version of this file under
'    the terms of any one of the MPL, the GPL or the LGPL.
'
'    ***** END LICENSE BLOCK *****


Imports System.Net.Sockets
Imports System.Text

Public Class ClientForm

    ' Connects to and pass connMessage to server.
    ' Waits for server reply.
    Public Function connect(ByVal connMessage As String) As String

        ' Enter server IP address here.
        Dim serverIPAddress As String = "10.0.0.101"
        ' Enter server listening port here.
        Dim serverListeningPort As Integer = 8000

        Dim returnStatus As String = ""
        Dim tcpClient As New System.Net.Sockets.TcpClient()
        Try
            ' Connect to server.
            tcpClient.Connect(serverIPAddress, serverListeningPort)
            Dim networkStream As NetworkStream = tcpClient.GetStream()

            ' Ensure valid network connection to server
            If networkStream.CanWrite And networkStream.CanRead Then
                ' Send connection message to server.
                Dim sendBytes As [Byte]() = Encoding.ASCII.GetBytes(connMessage)
                networkStream.Write(sendBytes, 0, sendBytes.Length)

                ' Read server response into a byte buffer.
                Dim recvBytes(1024) As Byte
                Dim recvData As StringBuilder = New StringBuilder()
                Do
                    networkStream.Read(recvBytes, 0, 1)
                    recvData.AppendFormat("{0}", Encoding.ASCII.GetString(recvBytes, 0, 1))
                Loop While networkStream.DataAvailable

                tcpClient.Close()
                returnStatus = recvData.ToString
            Else
                If Not networkStream.CanRead Then
                    messageBox.Text = "cannot not write" + vbNewLine + "data to this stream"
                    tcpClient.Close()
                ElseIf Not networkStream.CanWrite Then
                    messageBox.Text = "cannot read data" + vbNewLine + "from this stream"
                    tcpClient.Close()
                End If
            End If
        Catch e As Exception
            messageBox.Text = e.ToString()
        End Try
        Return returnStatus
    End Function

    ' Called when "Get Seat" button is clicked.
    Private Sub getSeat_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles getSeat.Click
        ' Connect to server with "get seat" message.
        Dim serverReturnStatus As String = connect("get seat")

        ' Determine action based on message returned from server.
        If StrComp(serverReturnStatus, "accept") = 0 Then
            messageBox.Text = "Connection Accepted" + vbNewLine + "You have a seat."
            getSeat.Enabled = False
            returnSeat.Enabled = True
        ElseIf StrComp(serverReturnStatus, "reject") = 0 Then
            messageBox.Text = "Connection Rejected" + vbNewLine + "No available seats."
            returnSeat.Enabled = False
            getSeat.Enabled = True
        ElseIf StrComp(serverReturnStatus, "unknown") = 0 Then
            messageBox.Text = "Unknown message" + vbNewLine + "returned from server."
        ElseIf StrComp(serverReturnStatus, "") = 0 Then
            messageBox.Text = "ERROR" + vbNewLine + "Could not connect" + vbNewLine + "to server."
        Else
            messageBox.Text = "ERROR" + vbNewLine + serverReturnStatus
        End If
    End Sub

    ' Called when "Return Seat" button is clicked.
    Private Sub returnSeat_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles returnSeat.Click
        ' Connect to server with "return seat" message.
        Dim serverReturnStatus As String = connect("return seat")

        ' Determine action based on message returned from server.
        If StrComp(serverReturnStatus, "returned") = 0 Then
            messageBox.Text = "Connection Accepted." + vbNewLine + "Seat Returned."
            getSeat.Enabled = True
            returnSeat.Enabled = False
        ElseIf StrComp(serverReturnStatus, "") = 0 Then
            messageBox.Text = "ERROR" + vbNewLine + "Unable to connect" + vbNewLine + "to server." + vbNewLine + "Seat not Returned."
        Else
            messageBox.Text = "ERROR" + vbNewLine + serverReturnStatus + vbNewLine + "Seat not Returned."
        End If
    End Sub

    ' Called when "Exit" button is clicked.
    Private Sub exitButton_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles exitButton.Click
        Me.Close()
    End Sub

    ' Called when form is closed.
    Private Sub Form1_FormClosing(ByVal sender As System.Object, ByVal e As System.Windows.Forms.FormClosingEventArgs) Handles MyBase.FormClosing
        ' Check it client currently has seat.
        If returnSeat.Enabled Then
            ' Connect to server with "return seat" message.
            Dim serverReturnStatus As String = connect("return seat")

            ' If there is a problem returning the seat, warn the user and let them decide
            ' whether or not to keep running the program or exit anyways.
            If StrComp(serverReturnStatus, "unknown") = 0 Or StrComp(serverReturnStatus, "") = 0 Then
                If MsgBox("Seat not returned. Exit Anyways?", MessageBoxButtons.YesNo, "ERROR") = Windows.Forms.DialogResult.No Then
                    e.Cancel = True
                End If
            End If
        End If
    End Sub

End Class
