'
' SmartDongle Visual Basic .NET Class Version 3.3
'
' MicroWorks, Inc.
' 2808 North Cole Road
' Boise, ID 83704
'
' http://www.smartdongle.com
'
' ***** BEGIN LICENSE BLOCK *****
'
' Version: MPL 1.1/GPL 2.0/LGPL 2.1
'
' The contents of this file are subject to the Mozilla Public License
' Version 1.1 (the "License"); you may not use this file except in
' compliance with the License. You may obtain a copy of the License at
' http://www.mozilla.org/MPL/
'
' Software distributed under the License is distributed on an "AS IS" basis,
' WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
' for the specific language governing rights and limitations under the
' License.
'
' The Original Code is MicroWorks, Inc. code.
'
' The Initial Developer of the Original Code is MicroWorks, Inc. Portions
' created by MicroWorks, Inc. are Copyright (C) 2006
' MicroWorks, Inc. All Rights Reserved.
'
' Contributor(s):
'
' Alternatively, the contents of this file may be used under the terms of
' either the GNU General Public License Version 2 or later (the "GPL"), or
' the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
' in which case the provisions of the GPL or the LGPL are applicable instead
' of those above. If you wish to allow use of your version of this file only
' under the terms of either the GPL or the LGPL, and not to allow others to
' use your version of this file under the terms of the MPL, indicate your
' decision by deleting the provisions above and replace them with the notice
' and other provisions required by the GPL or the LGPL. If you do not delete
' the provisions above, a recipient may use your version of this file under
' the terms of any one of the MPL, the GPL or the LGPL.
'
' ***** END LICENSE BLOCK *****


Imports System
Imports System.IO
Imports System.Threading
Imports System.Runtime.InteropServices

Public Class SmartDongle

    ' Valid SmartDongle serial numbers will always be this length
    Public Const CP2_SERIALNUM_LENGTH As Integer = 12

    ' SmartDongle GUID
    Private ClassGuid As Byte()

    Public Sub New()
        ' {5C98F9F9-C082-431A-A0AD-B32F70BD2689}
        ClassGuid = New Byte() {249, 249, 152, 92, 130, 192, 26, 67, 160, 173, 179, 47, 112, 189, 38, 137}
    End Sub

    Private Class Device

        Public Sub New()
            signature = New Byte(CY7C637XX.SIG_LENGTH - 1) {}
        End Sub

        Public signature As Byte()
        Public device As IntPtr
        Public options As Byte

        ' Driver version
        Public minor As Byte
        Public major As Byte

    End Class

    '
    ' Given the correct keys, will read SmartDongle memory into buffer.
    '
    Public Function Read( _
        ByVal P1 As Int64, _
        ByVal P2 As Int64, _
        ByVal address As Int16, _
        ByVal buffer() As Byte, _
        ByVal size As Integer) As SmartDnglError

        Dim dnglError As SmartDnglError
        Dim bufferSize As Integer = size
        Dim dongle As New Device

        ' An exception would occurr if size argument exceeded buffer length
        If buffer.Length < size Then
            Return SmartDnglError.UskErrBadArgument
        End If

        dnglError = OpenUsbDev(dongle)
        If dnglError <> SmartDnglError.UskOk Then
            Return dnglError
        End If

        ' Attempt to read user EEPROM space.
        ' On failure, attempt to disable dongle and return error.
        dnglError = Command.ReadEepromUserSpace( _
            dongle, P1, P2, address, buffer, bufferSize)

        If dnglError <> SmartDnglError.UskOk Then
            Command.DisableUsk(dongle)
        Else
            dnglError = Command.DisableUsk(dongle)
        End If

        Win32Methods.CloseHandle(dongle.device)

        Return dnglError
    End Function

    '
    ' Given the correct keys, will write buffer into SmartDongle memory.
    '
    Public Function Write( _
        ByVal P1 As Int64, _
        ByVal P2 As Int64, _
        ByVal address As Int16, _
        ByVal buffer() As Byte, _
        ByVal size As Integer) As SmartDnglError

        Dim dnglError As SmartDnglError
        Dim bufferSize As Integer = size
        Dim dongle As New Device


        dnglError = OpenUsbDev(dongle)
        If dnglError <> SmartDnglError.UskOk Then
            Return dnglError
        End If

        ' Attempt to write to user EEPROM space.
        ' On failure, attempt to disable dongle and return error.
        dnglError = Command.WriteEepromUserSpace( _
            P1, P2, address, buffer, bufferSize, dongle)

        If dnglError <> SmartDnglError.UskOk Then
            Command.DisableUsk(dongle)
        Else
            dnglError = Command.DisableUsk(dongle)
        End If

        'Win32Methods.CloseHandle()
        Win32Methods.CloseHandle(dongle.device)
        Return dnglError
    End Function

    '
    ' Reset, wait for SmartDongle to come ready.
    '
    Public Function Reset() As SmartDnglError

        Dim dnglError As SmartDnglError
        Dim dongle As New Device
        Dim retries As Integer


        dnglError = OpenUsbDev(dongle)
        If dnglError <> SmartDnglError.UskOk Then
            Return dnglError
        End If

        Command.Reset(dongle.device)
        Win32Methods.CloseHandle(dongle.device)

        Thread.Sleep(CY7C637XX.QTS_PLUG_N_PLAY_WAIT)

        For retries = 1 To CY7C637XX.QTS_READY_RETRIES

            dnglError = OpenUsbDev(dongle)
            If dnglError <> SmartDnglError.UskOk Then
                GoTo ContinueFor
            End If

            dnglError = IOCTL_Command.isDeviceReady(dongle.device, 0)
            If dnglError = SmartDnglError.UskOk Then
                Exit For
            End If

            Win32Methods.CloseHandle(dongle.device)

ContinueFor:
            Thread.Sleep(CY7C637XX.QTS_READY_WAIT)
        Next

        Win32Methods.CloseHandle(dongle.device)

        If dnglError <> SmartDnglError.UskOk Then
            Return SmartDnglError.UskResetComeready
        End If

        Return SmartDnglError.UskOk
    End Function

    '
    ' Read the SmartDongle onboard memory encryption mode.
    '
    Public Function ReadCryptMode(ByRef mode As Byte) As SmartDnglError

        Dim dnglError As SmartDnglError
        Dim dongle As New Device


        dnglError = OpenUsbDev(dongle)
        If dnglError <> SmartDnglError.UskOk Then
            Return dnglError
        End If

        dnglError = IOCTL_Command.GetMode(dongle.device, mode)

        Win32Methods.CloseHandle(dongle.device)

        Return dnglError
    End Function

    '
    ' Change the SmartDongle onboard memory encryption mode.
    '
    Public Function WriteCryptMode( _
        ByVal P1 As Int64, _
        ByVal P2 As Int64, _
        ByVal mode As Byte) As SmartDnglError

        Dim dnglError As SmartDnglError
        Dim dongle As New Device
        Dim currentMode As Byte
        Dim retries As Integer


        ' Mask out extraneous bits
        mode = _
            CByte( _
                mode And _
                (UserOption.QTS_OPTION_ON_CHIP_EPROM_PROTECT_DISABLE Or _
                UserOption.QTS_OPTION_ON_CHIP_EPROM_PROTECT_FAST))

        dnglError = OpenUsbDev(dongle)
        If dnglError <> SmartDnglError.UskOk Then
            Return dnglError
        End If

        ' Get fixed options
        dnglError = IOCTL_Command.GetOpt(dongle)
        If dnglError <> SmartDnglError.UskOk Then
            Win32Methods.CloseHandle(dongle.device)
            Return dnglError
        End If

        ' v1.50 SmartDongle must be reset if mode change
        If (dongle.options And CY7C637XX.QTS_OPTION_EXTENDED_SUPPORT2) = 0 And _
            (dongle.options And CY7C637XX.QTS_OPTION_EXTENDED_SUPPORT) <> 0 Then

            dnglError = IOCTL_Command.GetMode(dongle.device, currentMode)
            If dnglError <> SmartDnglError.UskOk Then
                Win32Methods.CloseHandle(dongle.device)
                Return dnglError
            End If

            ' Change mode only if current mode mismatches requested mode.
            If mode <> currentMode Then

                Command.Reset(dongle.device)
                Win32Methods.CloseHandle(dongle.device)

                Thread.Sleep(CY7C637XX.QTS_PLUG_N_PLAY_WAIT)

                For retries = 1 To CY7C637XX.QTS_READY_RETRIES

                    dnglError = OpenUsbDev(dongle)
                    If dnglError <> SmartDnglError.UskOk Then
                       GoTo ContinueFor
                    End If

                    dnglError = IOCTL_Command.isDeviceReady(dongle.device, 0)
                    If dnglError = SmartDnglError.UskOk Then
                       Exit For
                    End If

                    Win32Methods.CloseHandle(dongle.device)
ContinueFor:
                    Thread.Sleep(CY7C637XX.QTS_READY_WAIT)
                Next

                If dnglError <> SmartDnglError.UskOk Then
                    Win32Methods.CloseHandle(dongle.device)
                    Return SmartDnglError.UskResetComeready
                End If

                ' Requested mode same as current mode, return.
            Else
                Win32Methods.CloseHandle(dongle.device)
                Return SmartDnglError.UskOk
            End If
        End If

        dnglError = Command.EnableUsk(dongle, P1, P2)
        If dnglError <> SmartDnglError.UskOk Then
            Win32Methods.CloseHandle(dongle.device)
            Return dnglError
        End If

        dnglError = IOCTL_Command.SetMode(dongle.device, mode)

        Command.DisableUsk(dongle)

        Win32Methods.CloseHandle(dongle.device)

        Return dnglError
    End Function

    '
    ' SmartDongle Red LED on.
    '
    Public Function Red() As SmartDnglError

        Dim dnglError As SmartDnglError
        Dim dongle As New Device


        dnglError = OpenUsbDev(dongle)
        If dnglError <> SmartDnglError.UskOk Then
            Return dnglError
        End If

        dnglError = Command.Red(dongle.device)

        Win32Methods.CloseHandle(dongle.device)

        Return dnglError
    End Function

    '
    ' SmartDongle Green LED on.
    '
    Public Function Green() As SmartDnglError

        Dim dnglError As SmartDnglError
        Dim dongle As New Device


        dnglError = OpenUsbDev(dongle)
        If dnglError <> SmartDnglError.UskOk Then
            Return dnglError
        End If

        dnglError = Command.Green(dongle.device)

        Win32Methods.CloseHandle(dongle.device)

        Return dnglError
    End Function

    '
    ' SmartDongle LED off.
    '
    Public Function Off() As SmartDnglError

        Dim dnglError As SmartDnglError
        Dim dongle As New Device


        dnglError = OpenUsbDev(dongle)
        If dnglError <> SmartDnglError.UskOk Then
            Return dnglError
        End If

        dnglError = Command.Off(dongle.device)

        Win32Methods.CloseHandle(dongle.device)

        Return dnglError
    End Function

    '
    ' Retrieve SmartDongle serial number.
    '
    Public Function GetSerialNumber(ByRef sn() As Char) As SmartDnglError

        Dim dnglError As SmartDnglError
        Dim dongle As New Device


        dnglError = OpenUsbDev(dongle)
        If dnglError <> SmartDnglError.UskOk Then
            Return dnglError
        End If

        dnglError = IOCTL_Command.getSerialNumber(dongle.device, sn)

        Win32Methods.CloseHandle(dongle.device)

        Return dnglError
    End Function

    Public Class UserOption

        ' Three on-chip SmartDongle encrpytion modes to choose from.
        ' This provide protection for SmartDongle on-board memory only,
        ' and will not encrypt the data transferred over the bus.
        Public Const QTS_OPTION_ON_CHIP_EPROM_PROTECT_DISABLE As Byte = 1
        Public Const QTS_OPTION_ON_CHIP_EPROM_PROTECT_FAST As Byte = 2
        Public Const QTS_OPTION_ON_CHIP_EPROM_PROTECT_STRONG As Byte = 0
    End Class

      Private NotInheritable Class Win32Methods

        <StructLayout(LayoutKind.Sequential)> _
        Public Structure SP_DEVINFO_DATA
            Public cbSize As Int32
            Public Guid As Guid
            Public devInst As Int32
            Public Reserved As IntPtr
        End Structure

        <StructLayout(LayoutKind.Sequential)> _
        Public Structure SP_DEVICE_INTERFACE_DATA
            Public cbSize As Int32
            Public InterfaceClassGuid As Guid
            Public Flags As Int32
            Public Reserved As IntPtr
        End Structure

        <StructLayout(LayoutKind.Sequential, Pack:=1)> _
        Public Structure SP_DEVICE_INTERFACE_DETAIL_DATA
            Public cbSize As Int32
            ' dummy variable to get the struct size right
            Public DevicePath As Int16
        End Structure

        <DllImport( _
            "setupapi.dll", _
            EntryPoint:="SetupDiGetClassDevs", _
            SetLastError:=True, _
            CharSet:=CharSet.Auto, _
            ExactSpelling:=False, _
            CallingConvention:=CallingConvention.Winapi)> _
                Public Shared Function SetupDiGetClassDevs( _
                    ByVal ClassGuid() As Byte, _
                    ByVal Enumerator As IntPtr, _
                    ByVal hwndParent As IntPtr, _
                    ByVal Flags As Integer) As Integer
        End Function

        <DllImport( _
            "setupapi.dll", _
            EntryPoint:="SetupDiDestroyDeviceInfoList", _
            SetLastError:=True, _
            CharSet:=CharSet.Auto, _
            ExactSpelling:=False, _
            CallingConvention:=CallingConvention.Winapi)> _
                Public Shared Function SetupDiDestroyDeviceInfoList( _
                    ByVal DeviceInfoSet As Integer) As Boolean
        End Function


        <DllImport( _
            "setupapi.dll", _
            EntryPoint:="SetupDiEnumDeviceInterfaces", _
            SetLastError:=True, _
            CharSet:=CharSet.Auto, _
            ExactSpelling:=False, _
            CallingConvention:=CallingConvention.Winapi)> _
                Public Shared Function SetupDiEnumDeviceInterfaces( _
                    ByVal DeviceInfoSet As Integer, _
                    ByVal DeviceInfoData As IntPtr, _
                    ByVal ClassGuid() As Byte, _
                    ByVal MemberIndex As Integer, _
                    ByRef DeviceInterfaceData As Win32Methods.SP_DEVICE_INTERFACE_DATA) _
                       As Boolean
        End Function

        <DllImport( _
            "setupapi.dll", _
            EntryPoint:="SetupDiGetDeviceInterfaceDetail", _
            SetLastError:=True, _
            CharSet:=CharSet.Auto, _
            ExactSpelling:=False, _
            CallingConvention:=CallingConvention.Winapi)> _
        Public Overloads Shared Function SetupDiGetDeviceInterfaceDetail( _
                    ByVal DeviceInfoSet As Integer, _
                    ByRef DeviceInterfaceData As Win32Methods.SP_DEVICE_INTERFACE_DATA, _
                    ByVal DeviceInterfaceDetailData As Byte(), _
                    ByVal DeviceInterfaceDetailDataSize As Integer, _
                    ByRef RequiredSize As Integer, _
                    ByVal DeviceInfoData As IntPtr) As Boolean
        End Function

        <DllImport( _
            "setupapi.dll", _
            EntryPoint:="SetupDiGetDeviceInterfaceDetail", _
            SetLastError:=True, _
            CharSet:=CharSet.Auto, _
            ExactSpelling:=False, _
            CallingConvention:=CallingConvention.Winapi)> _
                Public Overloads Shared Function SetupDiGetDeviceInterfaceDetail( _
                    ByVal DeviceInfoSet As Integer, _
                    ByRef DeviceInterfaceData As Win32Methods.SP_DEVICE_INTERFACE_DATA, _
                    ByVal DeviceInterfaceDetailData As IntPtr, _
                    ByVal DeviceInterfaceDetailDataSize As Integer, _
                    ByRef RequiredSize As Integer, _
                    ByVal DeviceInfoData As IntPtr) As Boolean
        End Function

        <DllImport( _
            "kernel32.dll", _
            EntryPoint:="CreateFile", _
            SetLastError:=True, _
            CharSet:=CharSet.Auto, _
            ExactSpelling:=False, _
            CallingConvention:=CallingConvention.Winapi)> _
                Public Shared Function CreateFile( _
                    ByVal lpFileName() As Char, _
                    ByVal dwDesiredAccess As Integer, _
                    ByVal dwShareMode As Integer, _
                    ByVal lpSecurityAttributes As IntPtr, _
                    ByVal dwCreationDisposition As Integer, _
                    ByVal dwFlagsAndAttributes As Integer, _
                    ByVal hTemplateFile As IntPtr) As IntPtr
        End Function

        <DllImport( _
            "kernel32.dll", _
            EntryPoint:="CloseHandle", _
            SetLastError:=True, _
            CharSet:=CharSet.Auto, _
            ExactSpelling:=False, _
            CallingConvention:=CallingConvention.Winapi)> _
                Public Shared Function CloseHandle( _
                    ByVal hObject As IntPtr) As Integer
        End Function

        <DllImport( _
            "kernel32.dll", _
            EntryPoint:="DeviceIoControl", _
            SetLastError:=True, _
            CharSet:=CharSet.Auto, _
            ExactSpelling:=False, _
            CallingConvention:=CallingConvention.Winapi)> _
                Public Overloads Shared Function DeviceIoControl( _
                    ByVal hDevice As IntPtr, _
                    ByVal dwIoControlCode As Int32, _
                    ByVal lpInBuffer As IntPtr, _
                    ByVal nInBufferSize As Integer, _
                    ByVal lpOutBuffer As IntPtr, _
                    ByVal nOutBufferSize As Integer, _
                    ByRef lpBytesReturned As Integer, _
                    ByVal lpOverlapped As IntPtr) As Boolean
        End Function

        <DllImport( _
            "kernel32.dll", _
            EntryPoint:="DeviceIoControl", _
            SetLastError:=True, _
            CharSet:=CharSet.Auto, _
            ExactSpelling:=False, _
            CallingConvention:=CallingConvention.Winapi)> _
                Public Overloads Shared Function DeviceIoControl( _
                    ByVal hDevice As IntPtr, _
                    ByVal dwIoControlCode As Int32, _
                    ByVal lpInBuffer() As Byte, _
                    ByVal nInBufferSize As Integer, _
                    ByVal lpOutBuffer As IntPtr, _
                    ByVal nOutBufferSize As Integer, _
                    ByRef lpBytesReturned As Integer, _
                    ByVal lpOverlapped As IntPtr) As Boolean
        End Function

        <DllImport( _
            "kernel32.dll", _
            EntryPoint:="DeviceIoControl", _
            SetLastError:=True, _
            CharSet:=CharSet.Auto, _
            ExactSpelling:=False, _
            CallingConvention:=CallingConvention.StdCall)> _
                Public Overloads Shared Function DeviceIoControl( _
                    ByVal hDevice As IntPtr, _
                    ByVal dwIoControlCode As Int32, _
                    ByVal lpInBuffer As IntPtr, _
                    ByVal nInBufferSize As Integer, _
                    ByVal lpOutBuffer() As Byte, _
                    ByVal nOutBufferSize As Integer, _
                    ByRef lpBytesReturned As Integer, _
                    ByVal lpOverlapped As IntPtr) As Boolean
        End Function

        <DllImport( _
            "kernel32.dll", _
            EntryPoint:="DeviceIoControl", _
            SetLastError:=True, _
            CharSet:=CharSet.Auto, _
            ExactSpelling:=False, _
            CallingConvention:=CallingConvention.Winapi)> _
                Public Overloads Shared Function DeviceIoControl( _
                    ByVal hDevice As IntPtr, _
                    ByVal dwIoControlCode As Int32, _
                    ByVal lpInBuffer() As Byte, _
                    ByVal nInBufferSize As Integer, _
                    ByVal lpOutBuffer() As Byte, _
                    ByVal nOutBufferSize As Integer, _
                    ByRef lpBytesReturned As Integer, _
                    ByVal lpOverlapped As IntPtr) As Boolean
        End Function

    End Class

    Private NotInheritable Class Command

        '
        ' Read from an unlocked SmartDongle, see sd_uskerr.h for returned errors.
        '
        Public Shared Function ReadEepromUserSpace( _
            ByVal dongle As Device, _
            ByVal P1 As Int64, _
            ByVal P2 As Int64, _
            ByVal address As Int16, _
            ByVal buffer() As Byte, _
            ByRef size As Integer) As SmartDnglError

            Dim flags As Byte
            Dim dnglError As SmartDnglError = SmartDnglError.UskOk
            Dim i, j, retries As Integer
            Dim offset As Int16
            Dim address_bytes(CY7C637XX.ADDR_LENGTH) As Byte
            Dim out_buffer(8 * CY7C637XX.EEPROMPAGESIZE) As Byte
            Dim page0 As Byte
            Dim firstPrefetch As Boolean
            Dim singlePrefetch As Boolean
            Dim prefetches As Integer
            Dim prefetchSize As Integer
            Dim to_load As Integer
            Dim bytesReturned As Integer = 0


            dnglError = IOCTL_Command.GetOpt(dongle)
            If dnglError <> SmartDnglError.UskOk Then
                Return dnglError
            End If

            prefetches = 1
            prefetchSize = 0
            singlePrefetch = True
            firstPrefetch = True

            ' User may request address that is inside a page.
            ' Get the offset from the first page.
            page0 = CByte(address And &HF)

            ' Mask out lower address nibble
            address = CShort(address And &HFFF0)

            ' Int16 address to big endian byte array
            address_bytes = BitConverter.GetBytes(address)

            to_load = size
            size = 0

            dnglError =  _
                IOCTL_Command.waitForBusy( dongle.device, CY7C637XX.QTSBUSYWAIT)
            If dnglError <> SmartDnglError.UskOk Then
                Return dnglError
            End If

            i = 0
            While i < prefetches

                offset = CType(i * prefetchSize + address, Int16)

                ' Must obtain new key for each read transaction if strict
                ' access is required by SmartDongle.
                ' Extended functionality must be supported in firmware.
                If (dongle.options And CY7C637XX.QTS_OPTION_STRICT_ONLY) > 0 Then
                    dnglError = EnableUsk(dongle, P1, P2)
                Else
                    If firstPrefetch Then
                       dnglError = EnableUsk(dongle, P1, P2)
                    End If
                End If

                ' Return on error
                If dnglError <> SmartDnglError.UskOk Then
                    Return dnglError
                End If

                For retries = 0 To CY7C637XX.QTS_READ_RETRIES - 1
                    ' Request prefetch starting at offset
                    dnglError = IOCTL_Command.requestEpromPrefetch( _
                       dongle.device, offset)
                    If dnglError <> SmartDnglError.UskOk Then
                       Exit For
                    End If

                    ' Return error status if request failed
                    dnglError = IOCTL_Command.waitForPageRead( _
                       dongle.device, CY7C637XX.QTSPAGEWAIT, flags)
                    If dnglError <> SmartDnglError.UskOk Then
                       Exit For
                    End If

                    ' Don't retry on an address range error
                    If (flags And CY7C637XX.QTS_READ_ERROR_MASK) > 0 Then

                       dnglError = SmartDnglError.UskErrEepromRead

                       If (flags And (CY7C637XX.QTSRDDENY Or CY7C637XX.QTSRDEEPROMERROR)) > 0 _
                       Then
                          GoTo ContinueFor
                       Else
                          Exit For
                       End If
                    End If

                    ' Read buffered pages
                    dnglError = IOCTL_Command.readEpromPrefetch( _
                       out_buffer, bytesReturned, dongle)

                    Exit For
ContinueFor:
                Next retries

                ' Return on error
                If dnglError <> SmartDnglError.UskOk Then
                    Return dnglError
                End If

                ' Use the number of bytes returned calculated the number of
                ' prefetches needed.
                ' Calculate using the offset into the first EPROM page.
                If firstPrefetch Then
                    '
                    ' Sanity check, default to single page of prefetch.
                    ' the 1.5 returns 1024 but has only 8 bytes of data.
                    '
                    If bytesReturned < CY7C637XX.EEPROMPAGESIZE Or bytesReturned > CY7C637XX.QTS_MAXPREFETCH Then
                       prefetchSize = CY7C637XX.EEPROMPAGESIZE
                    Else
                       prefetchSize = bytesReturned
                    End If

                    prefetches = CInt(System.Math.Floor((to_load + page0) / prefetchSize))

                    ' Another prefetch for remainder
                    If (to_load + page0) Mod prefetchSize > 0 Then
                       prefetches += 1
                    End If

                    ' A single prefetch is read with regard to page 0 offset and
                    ' data length remainder.
                    If prefetches > 1 Then
                       singlePrefetch = False
                    Else
                       singlePrefetch = True
                    End If
                End If

                ' Process prefetches.
                If singlePrefetch Then
                    For j = 0 To to_load - 1
                       buffer(j) = out_buffer((j + page0))
                    Next j

                    size = to_load
                Else
                    If to_load + page0 >= prefetchSize Then
                       If firstPrefetch Then
                          For j = 0 To (prefetchSize - page0) - 1
                             buffer(j) = out_buffer((j + page0))
                          Next j

                          size += prefetchSize - page0
                          to_load -= prefetchSize - page0
                       Else
                          For j = 0 To prefetchSize - 1
                             buffer((j + size)) = out_buffer(j)
                          Next j

                          size += prefetchSize
                          to_load -= prefetchSize
                       End If
                    Else

                       ' Load remainder.
                       ' A "first prefetch" here would be a "single prefetch"
                       For j = 0 To to_load - 1
                          buffer((j + size)) = out_buffer(j)
                       Next j

                       size += to_load
                       to_load = 0
                    End If
                End If
                firstPrefetch = False

                i += 1
            End While

            Return SmartDnglError.UskOk
        End Function

        '
        ' Write to an unlocked SmartDongle, see sd_uskerr.h for returned errors.
        '
        Public Shared Function WriteEepromUserSpace( _
            ByVal P1 As Int64, _
            ByVal P2 As Int64, _
            ByVal address As Int16, _
            ByVal buffer() As Byte, _
            ByRef size As Integer, _
            ByVal dongle As Device) As SmartDnglError

            Dim flags As Byte
            Dim dnglError As SmartDnglError = SmartDnglError.UskOk

            Dim to_send, sent, received As Integer
            Dim i, j, retries, pages As Integer

            Dim offset As Int16
            Dim address_bytes(CY7C637XX.ADDR_LENGTH) As Byte
            Dim page(CY7C637XX.EEPROMPAGESIZE) As Byte
            Dim page0 As Byte
            Dim firstPage As Boolean
            Dim singlePage As Boolean


            dnglError = IOCTL_Command.GetOpt(dongle)
            If dnglError <> SmartDnglError.UskOk Then
                Return dnglError
            End If

            ' User may request address that is inside a page.
            ' Get the offset from the first page.
            page0 = CByte(address And &HF)

            ' Mask out lower address nibble
            address = CShort(address And &HFFF0)

            ' Int16 address to big endian byte array
            address_bytes = BitConverter.GetBytes(address)

            to_send = size
            size = 0
            firstPage = True
            singlePage = True

            ' Calculate number of write pages with respect to any offset into the
            ' first EPROM page.
            pages = CInt(System.Math.Floor((to_send + CInt(page0)) / CY7C637XX.EEPROMPAGESIZE))
            If (to_send + CInt(page0)) Mod CY7C637XX.EEPROMPAGESIZE > 0 Then
                pages += 1
            End If

            ' If a single page is written,
            ' page 0 offset and the data length remainder is regarded.
            If pages > 1 Then
                singlePage = False
            End If

            For i = 0 To pages - 1

                offset = CType(i * CY7C637XX.EEPROMPAGESIZE + address, Int16)

                ' First and last page - single page.
                If singlePage Then

                    ' Retrieve current EPROM page if entire page not overwritten.
                    If to_send < CY7C637XX.EEPROMPAGESIZE Then

                       received = CY7C637XX.EEPROMPAGESIZE
                       dnglError = ReadEepromUserSpace( _
                          dongle, P1, P2, offset, page, received)

                       ' Skip size check if IOCTL error
                       If dnglError = SmartDnglError.UskOk Then
                          If received <> CY7C637XX.EEPROMPAGESIZE Then
                             dnglError = SmartDnglError.UskErrEepromRead
                          End If
                       End If
                    End If

                    For j = 0 To to_send - 1
                       page((j + CInt(page0))) = buffer(j)
                    Next j

                    sent = to_send

                Else

                    ' Process all pages but the last page here.
                    If (i = pages - 1) = False Then

                       ' Address an offset into first page,
                       ' then fetch and overwrite tail end of first page.
                       If firstPage And page0 > 0 Then

                          received = CY7C637XX.EEPROMPAGESIZE
                          dnglError = ReadEepromUserSpace( _
                             dongle, P1, P2, offset, page, received)

                          ' Skip size check if IOCTL error
                          If dnglError = SmartDnglError.UskOk Then
                             If received <> CY7C637XX.EEPROMPAGESIZE Then
                                dnglError = SmartDnglError.UskErrEepromRead
                             End If
                          End If

                          For j = 0 To (CY7C637XX.EEPROMPAGESIZE - CInt(page0)) - 1
                             page((j + CInt(page0))) = buffer(j)
                          Next j

                          sent = CY7C637XX.EEPROMPAGESIZE - CInt(page0)
                       Else

                          ' No offset, simply overwrite entire buffer.
                          For j = 0 To CY7C637XX.EEPROMPAGESIZE - 1
                             page(j) = buffer((j + size))
                          Next j

                          sent = CY7C637XX.EEPROMPAGESIZE
                       End If

                    Else

                       ' Lase Page
                       '
                       ' Retrieve current EPROM page if
                       ' entire page not to be overwritten.
                       If to_send < CY7C637XX.EEPROMPAGESIZE Then

                          received = CY7C637XX.EEPROMPAGESIZE
                          dnglError = ReadEepromUserSpace( _
                             dongle, P1, P2, offset, page, received)

                          ' Skip size check if IOCTL error
                          If dnglError = SmartDnglError.UskOk Then
                             If received <> CY7C637XX.EEPROMPAGESIZE Then
                                dnglError = SmartDnglError.UskErrEepromRead
                             End If
                          End If
                       End If

                       For j = 0 To to_send - 1
                          page(j) = buffer((j + size))
                       Next j

                       sent = to_send
                    End If
                End If

                ' Return on error
                If dnglError <> SmartDnglError.UskOk Then
                    Return dnglError
                End If

                ' Must obtain new key for each write transaction if strict
                ' access was requested or if SmartDongle requires strict access.
                ' Extended functionality must be supported in firmware.
                If (dongle.options And CY7C637XX.QTS_OPTION_STRICT_ONLY) > 0 Then
                    dnglError = EnableUsk(dongle, P1, P2)
                Else
                    If firstPage Then
                       dnglError = EnableUsk(dongle, P1, P2)
                    End If
                End If

                ' Return on error
                If dnglError <> SmartDnglError.UskOk Then
                    Return dnglError
                End If

                For retries = 0 To CY7C637XX.QTS_WRITE_RETRIES - 1

                    ' Write the page
                    dnglError = IOCTL_Command.writeEpromPage(offset, page, dongle)
                    If dnglError <> SmartDnglError.UskOk Then
                       Exit For
                    End If

                    dnglError = IOCTL_Command.waitForBusy( _
                       dongle.device, CY7C637XX.QTSBUSYWAIT)
                    If dnglError <> SmartDnglError.UskOk Then
                       Exit For
                    End If

                    ' Any errors writing page to EEPROM?
                    dnglError = IOCTL_Command.readPageFlags(dongle.device, flags)
                    If dnglError <> SmartDnglError.UskOk Then
                       Exit For
                    End If

                    ' Don't retry if an address range error
                    flags = CByte(flags And CY7C637XX.QTS_WRITE_ERROR_MASK)
                    If flags > 0 Then

                       dnglError = SmartDnglError.UskErrEepromWrite

                       If (flags And (CY7C637XX.QTSWRDENY Or CY7C637XX.QTSWREEPROMERROR)) > 0 _
                       Then
                          ' Ignore error
                          IOCTL_Command.waitForTwc( _
                             dongle.device, CY7C637XX.QTSTWCWAIT)
                          GoTo ContinueFor
                       Else
                          Exit For
                       End If
                    End If

                    ' Wait here for Twc (Time Write cycle) to reset.
                    dnglError = IOCTL_Command.waitForTwc( _
                       dongle.device, CY7C637XX.QTSTWCWAIT)

                    Exit For
ContinueFor:
                Next retries

                ' return on error
                If dnglError <> SmartDnglError.UskOk Then
                    Return dnglError
                End If

                to_send -= sent
                size += sent

                firstPage = False
            Next i

            Return SmartDnglError.UskOk
        End Function

        '
        ' Enable Dongle by stepping through the required States and using Keys.
        '
        Public Shared Function EnableUsk( _
            ByVal dongle As Device, _
            ByVal P1 As Int64, _
            ByVal P2 As Int64) As SmartDnglError

            Dim challenge As Int64
            Dim X As Int64
            Dim state As Byte
            Dim dnglError As SmartDnglError


            dnglError = IOCTL_Command.GetOpt(dongle)
            If dnglError <> SmartDnglError.UskOk Then
                Win32Methods.CloseHandle(dongle.device)
                Return dnglError
            End If

            IOCTL_Command.GetDriverVersion(dongle)

            ' Reset key state machine here for all but >=v1.50 SmartDongles
            If (dongle.options And CY7C637XX.QTS_OPTION_EXTENDED_SUPPORT) = 0 Then

                dnglError = IOCTL_Command.resetKeyState(dongle.device)
                If dnglError <> SmartDnglError.UskOk Then
                    Return dnglError
                End If

                dnglError = IOCTL_Command.waitForKeyState( _
                    dongle.device, CY7C637XX.QTSINITIAL, CY7C637XX.QTSSTATEWAIT)
                If dnglError <> SmartDnglError.UskOk Then

                    ' Was there a fatal EPROM read error?
                    ' Ignore readPageFlags IOCTL error
                    If IOCTL_Command.readPageFlags(dongle.device, state) = SmartDnglError.UskOk Then
                       If (state And CY7C637XX.QTS_READ_ERROR_MASK) > 0 Then
                          Return SmartDnglError.UskErrEepromRead
                       End If
                    End If

                    Return dnglError
                End If
            End If

            ' Request the dongle create challenge.
            dnglError = IOCTL_Command.requestHostKey(dongle.device)
            If dnglError <> SmartDnglError.UskOk Then
                Return dnglError
            End If

            ' Dongle's internally generated number to challenge host ready?
            dnglError = IOCTL_Command.waitForKeyState( _
                dongle.device, CY7C637XX.QTSREPLY, CY7C637XX.QTSSTATEWAIT)
            If dnglError <> SmartDnglError.UskOk Then

                ' Was there a fatal EPROM read error?
                ' Ignore readPageFlags IOCTL error here
                If IOCTL_Command.readPageFlags(dongle.device, state) = SmartDnglError.UskOk Then
                    If (state And CY7C637XX.QTS_READ_ERROR_MASK) > 0 Then
                       Return SmartDnglError.UskErrEepromRead
                    End If
                End If

                Return dnglError
            End If

            ' Get the SmartDongle's challenge.
            dnglError = IOCTL_Command.getKey(dongle, challenge)
            If dnglError <> SmartDnglError.UskOk Then
                Return dnglError
            End If

            X = Multiply(challenge, P1)
            X = Add(X, P2)

            ' Reply to challenge
            dnglError = IOCTL_Command.sendKey(dongle, X)
            If dnglError <> SmartDnglError.UskOk Then
                Return dnglError
            End If

            ' When key state sequences back to 0 (QTSINITIAL),
            ' the host response has been verified.
            ' The security state should be 1 (QTSENABLED) if the
            ' host reply is correct.
            dnglError = IOCTL_Command.waitForKeyState( _
                dongle.device, CY7C637XX.QTSINITIAL, CY7C637XX.QTSSTATEWAIT)
            If dnglError <> SmartDnglError.UskOk Then

                ' Was there a fatal EPROM read error?
                ' Ignore readPageFlags IOCTL error here
                If IOCTL_Command.readPageFlags(dongle.device, state) = SmartDnglError.UskOk Then
                    If (state And CY7C637XX.QTS_READ_ERROR_MASK) > 0 Then
                       Return SmartDnglError.UskErrEepromRead
                    End If
                End If

                Return dnglError
            End If

            ' Is the host verified, dongle ENABLED?
            dnglError = IOCTL_Command.readSecurityState(dongle.device, state)
            If dnglError <> SmartDnglError.UskOk Then
                Return dnglError
            End If
            If state <> CY7C637XX.QTSENABLED Then
                Return SmartDnglError.UskErrHostValidate
            End If

            '  Force a v1.50 or greater SmartDongle to identify itself.
            If (dongle.options And CY7C637XX.QTS_OPTION_EXTENDED_SUPPORT) > 0 Then
                dnglError = ValidateUsk(dongle, P1, P2)
                If dnglError <> SmartDnglError.UskOk Then
                    IOCTL_Command.resetKeyState(dongle.device)
                End If
            End If

            Return dnglError
        End Function

        '
        ' Key state and Security state to 0 locks SmartDongle
        '
        Public Shared Function DisableUsk(ByVal dongle As Device) As SmartDnglError
            Return IOCTL_Command.resetKeyState(dongle.device)
        End Function

        '
        ' For SmartDongle v1.50, force a SmartDongle to reply to
        ' a host challenge. Useful if a user does not write any
        ' secrets to SmartDongle's user memory and simply checks
        ' for presence of a SmartDongle. An attacker could spoof
        ' a protected application program by using a modified
        ' driver or usk.dll
        '
        ' SmartDongle must be enabled to support a host challenge.
        Public Shared Function ValidateUsk( _
            ByVal dongle As Device, _
            ByVal P1 As Int64, _
            ByVal P2 As Int64) As SmartDnglError

            Dim challenge, reply, expected_reply As Int64
            Dim flags As Byte
            Dim dnglError As SmartDnglError
            Dim InDate As Date


            ' Make a 64 bit challenge
            InDate = TimeOfDay
            Dim rand As New Random(CInt(InDate.TimeOfDay.TotalSeconds()))
            challenge = CType(rand.Next(), Int64)
            challenge = challenge Or Multiply(CType(rand.Next(), Int64), &H10000)
            challenge = challenge Or Multiply(CType(rand.Next(), Int64), &H100000000)

            ' High bit must be set
            challenge = challenge Or &H8000000000000000

            ' Calculate expected reply from SmartDongle
            expected_reply = Multiply(challenge, P1)
            expected_reply = Add(expected_reply, P2)

            ' Send challenge.
            ' Assuming key state was set to 0 at termination of
            ' SmartDongle enable sequence.
            dnglError = IOCTL_Command.sendChallenge(dongle, challenge)
            If dnglError <> SmartDnglError.UskOk Then
                Return dnglError
            End If
            ' Wait for SmartDongle's reply to host challenge
            dnglError = IOCTL_Command.waitForKeyState( _
                dongle.device, CY7C637XX.QTSREPLY, CY7C637XX.QTSSTATEWAIT)
            If dnglError <> SmartDnglError.UskOk Then

                ' Was there a fatal EPROM read error?
                ' Ignore readPageFlags IOCTL error here
                If IOCTL_Command.readPageFlags(dongle.device, flags) = SmartDnglError.UskOk Then
                    If (flags And CY7C637XX.QTS_READ_ERROR_MASK) > 0 Then
                       Return SmartDnglError.UskErrEepromRead
                    End If
                End If
                Return dnglError
            End If

            ' Get the SmartDongle's reply to challenge.
            dnglError = IOCTL_Command.getKey(dongle, reply)
            If dnglError <> SmartDnglError.UskOk Then
                Return dnglError
            End If
            If expected_reply <> reply Then
                Return SmartDnglError.UskErrUskValidate
            End If
            Return SmartDnglError.UskOk
        End Function

        ' Reset SmartDongle
        Public Shared Function Reset(ByVal device As IntPtr) As SmartDnglError
            Return IOCTL_Command.SoftReset(device)
        End Function


        ' Output port control for LED
        Public Shared Function Green(ByVal device As IntPtr) As SmartDnglError
            Return IOCTL_Command.writePort( _
                device, CY7C637XX.QTSLEDPORT, CY7C637XX.QTSGREEN)
        End Function

        Public Shared Function Red(ByVal device As IntPtr) As SmartDnglError
            Return IOCTL_Command.writePort( _
                device, CY7C637XX.QTSLEDPORT, CY7C637XX.QTSRED)
        End Function

        Public Shared Function Off(ByVal device As IntPtr) As SmartDnglError
            Return IOCTL_Command.writePort( _
                device, CY7C637XX.QTSLEDPORT, CY7C637XX.QTSLEDOFF)
        End Function

    End Class

    Private NotInheritable Class CY7C637XX

        ' User memory page size
        Public Const EEPROMPAGESIZE As Integer = 16

        ' Number of prefetch pages in an enabled SmartDongle
        Public Const CP2_PAGE_PREFETCH_ENABLED As Integer = 8

        ' Length of serial number USB descriptor string
        Public Const CP2_SER_NUMB_STRING_LENGTH As Integer = 2 + (CP2_SERIALNUM_LENGTH * 2)

        ' usb100.h
        Public Const USB_STRING_DESCRIPTOR_TYPE As Integer = 3

        ' 32 bit signature for read/write access
        Public Const SIG_LENGTH As Integer = 4

        ' 16 bit address
        Public Const ADDR_LENGTH As Integer = 2

        ' 64 bit keys
        Public Const SECRETS_KEY_LEN As Integer = 8

        '
        ' Fixed options on SmartDongle queried by GetOpt()
        '

        ' SmartDongle v1.70 or above, extended functionality
        Public Const QTS_OPTION_EXTENDED_SUPPORT2 As Byte = &H20

        ' SmartDongle v1.50 or above, extended functionality
        Public Const QTS_OPTION_EXTENDED_SUPPORT As Byte = &H80

        ' SmartDongle requires a new key for each access
        Public Const QTS_OPTION_STRICT_ONLY As Byte = &H40

        ' Firmware option bitmap
        Public Const QTS_FIRMWARE_EXTENDED_SUPPORT2 As Byte = &H40
        Public Const QTS_FIRMWARE_EXTENDED_SUPPORT As Byte = &H08
        Public Const QTS_FIRMWARE_STRICT_ONLY As Byte = &H4

        Public Const QTS_READ_PAGE_READY_MASK As Byte = _
            QTSRDPAGEVALID Or QTSRDADDRESSRANGE Or QTSRDDENY Or QTSRDEEPROMERROR

        Public Const QTS_WRITE_ERROR_MASK As Byte = _
             QTSWRADDRESSRANGE Or QTSWRDENY Or QTSWREEPROMERROR

        Public Const QTS_READ_ERROR_MASK As Byte = QTSRDEEPROMERROR

        '
        ' sd_helper.h
        '
        Public Const QTS_IOCTL_RETRYS As Integer = 5
        Public Const QTS_IOCTL_FAIL_WAIT As Integer = 100
        Public Const QTS_IOCTL_BUFFSIZE As Integer = 1024

        Public Const QTS_READ_RETRIES As Integer = 3
        Public Const QTS_WRITE_RETRIES As Integer = 3
        Public Const QTS_SCAN_RETRIES As Integer = 1

        Public Const QTS_READY_RETRIES As Integer = 8
        Public Const QTS_READY_WAIT As Integer = 250
        Public Const QTS_PLUG_N_PLAY_WAIT As Integer = 1000

        Public Const QTS_MAXPREFETCH As Integer = 128

        '
        ' Various timeouts (in mSec).
        '

        ' Max time period to wait for an EEPROM page
        Public Const QTSPAGEWAIT As Integer = 200

        ' Max time period to wait for a state change
        Public Const QTSSTATEWAIT As Integer = 250

        ' Max time period to wait for Time Write Cycle.
        Public Const QTSTWCWAIT As Integer = 200

        ' Max time period to wait for task_loop
        Public Const QTSBUSYWAIT As Integer = 500

        ' Each USB low speed packet is 8 bytes
        Public Const QTSPACKETSIZE As Integer = 8

        Public Const QTSMAXBUFFERLENGTH As Integer = 255

        ' Page flag bits: IOCTL_QTSDONGLE_GET_PAGE_FLAGS
        Public Const QTSRDPAGEVALID As Byte = &H1
        Public Const QTSWRPAGEVALID As Byte = &H2

        ' EEPROM read page errors
        Public Const QTSRDADDRESSRANGE As Byte = &H4
        Public Const QTSRDDENY As Byte = &H8
        Public Const QTSRDEEPROMERROR As Byte = &H10  ' SPI EPROM read timeout

        ' EEPROM write page errors
        Public Const QTSWRADDRESSRANGE As Byte = &H20
        Public Const QTSWRDENY As Byte = &H40
        Public Const QTSWREEPROMERROR As Byte = &H80 ' SPI EPROM write timeout

        '
        ' EEPROM data error bits: IOCTL_QTSDONGLE_GET_EEPROM_ERROR_FLAG
        '

        ' EEPROM SPI access timeout
        Public Const QTSRDTIMEOUT As Byte = &H1

        ' EEPROM SPI access timeout and data miscompare
        Public Const QTSWRTIMEOUT As Byte = &H10
        Public Const QTSWRDATA As Byte = &H20

        ' The status of the Time Write Cycle register.
        ' If the register indicates that a write cycle is
        ' active, an EEPROM read or write is denied.
        ' IOCTL_QTSDONGLE_GET_TWC_FLAG
        Public Const QTSTWCACTIVE As Byte = &HFF
        Public Const QTSTWCINACTIVE As Byte = &H0

        ' Task loop activity flag for IOCTL_QTSDONGLE_GET_BUSY_FLAG
        Public Const QTSREADY As Byte = &H0
        Public Const QTSBUSY As Byte = &HFF

        ' The 8 bit ouput port address, the CY7C63221 has P0.0 - P0.8, and P2.0, P2.1
        ' (XTALIN, XTALOUT disabled)
        Public Const QTSPORT0 As Byte = &H0
        Public Const QTSPORT1 As Byte = &H1
        Public Const QTSPORT2 As Byte = &H2

        ' The 8 bit output port LED drive is active low, using lines 0 and 1
        Public Const QTSLEDPORT As Byte = QTSPORT0
        Public Const QTSLEDOFF As Byte = &HFF
        Public Const QTSGREEN As Byte = &HFE
        Public Const QTSRED As Byte = &HFD

        ' Security states
        Public Const QTSDISABLED As Byte = &H0
        Public Const QTSENABLED As Byte = &H1

        ' Key states for host/dongle authentication handshake
        Public Const QTSINITIAL As Byte = &H0
        Public Const QTSREPLY As Byte = &H1

        ' Bitmap of SmartDongle status byte.
        Public Const QTSKEYSTATEBIT As Byte = &H80
        Public Const QTSKEYINITIAL As Byte = &H0
        Public Const QTSKEYREPLY As Byte = &H80

        Public Const QTSTWCSTATEBIT As Byte = &H4
        Public Const QTSTWCBUSY As Byte = &H4
        Public Const QTSTWCDONE As Byte = &H0

        Public Const QTSBUSYSTATEBIT As Byte = &H20
        Public Const QTSISBUSY As Byte = &H20
        Public Const QTSNOTBUSY As Byte = &H0

        ' Read SmartDongle's extended functionality bits are mapped to
        ' EEPROM error bits.
        '
        ' This bit will be set in v1.50. Indicates that new
        ' functionality is supported.
        Public Const QTSEXTSUPPORT As Byte = &H8

        ' If this bit is set in v1.50, strict access only
        Public Const QTSSTRICTONLY As Byte = &H4


        Public Shared Function SignatureBytesFromKey( _
            ByVal key As Int64) As Byte()

            Dim tmp As Byte
            Dim data() As Byte
            Dim signature(CY7C637XX.SIG_LENGTH - 1) As Byte
            Dim i As Integer


            data = BitConverter.GetBytes(key)

            ' SmartDongle is expecting two big-endian shorts
            tmp = data(5)
            data(5) = data(7)
            data(7) = tmp

            tmp = data(4)
            data(4) = data(6)
            data(6) = tmp

            For i = 0 To CY7C637XX.SIG_LENGTH - 1
                signature(i) = data(i + 4)
            Next i

            Return signature

        End Function

        Public Shared Function BytesFromKey(ByVal key As Int64) As Byte()

            Dim i As Integer, j As Integer
            Dim data() As Byte
            Dim tmp As Byte


            data = BitConverter.GetBytes(key)

            j = 7

            For i = 0 To 3
                tmp = data(i)
                data(i) = data(j)
                data(j) = tmp
                j = j - 1
            Next i

            Return data
        End Function


    End Class

    Private NotInheritable Class IOCTL_Command

        '
        ' Send a challenge to SmartDongle.
        '
        Public Shared Function sendChallenge( _
            ByVal dongle As Device, _
            ByVal challenge As Int64) As SmartDnglError

            Dim i As Integer
            Dim dnglError As SmartDnglError
            Dim bytesReturned As Integer
            Dim status As Boolean = False
            Dim buffer As IntPtr


            ' For v1.50 or greater SmartDongles only
            If (dongle.options And CY7C637XX.QTS_OPTION_EXTENDED_SUPPORT) = 0 Then
                Return SmartDnglError.UskErrRequiresNewerSmartDongleVersion
            End If

            buffer = Marshal.AllocHGlobal(Marshal.SizeOf(challenge))
            Marshal.WriteInt64(buffer, CType(challenge, Int64))

            ' Wait for task_loop processing to finish
            dnglError = waitForBusy(dongle.device, CY7C637XX.QTSBUSYWAIT)
            If dnglError <> SmartDnglError.UskOk Then
                Return dnglError
            End If

            ' Driver does not queue requests, retry on fail
            For i = 0 To CY7C637XX.QTS_IOCTL_RETRYS - 1
                status = Win32Methods.DeviceIoControl( _
                dongle.device, _
                IOCTL_QTSDONGLE_SEND_KEY(), _
                buffer, _
                CY7C637XX.SECRETS_KEY_LEN, _
                IntPtr.Zero, _
                0, _
                bytesReturned, _
                IntPtr.Zero)

                If status Then
                    Exit For
                End If

                Thread.Sleep(CY7C637XX.QTS_IOCTL_FAIL_WAIT)
            Next i

            Marshal.FreeHGlobal(buffer)

            If status = False Then
                Return SmartDnglError.UskErrIoctlSendKey
            End If

            Return SmartDnglError.UskOk
        End Function


        '
        ' Send a Reply calculated from SmartDongle's Challenge.
        '
        Public Shared Function sendKey( _
            ByRef dongle As Device, _
            ByVal reply As Int64) As SmartDnglError

            Dim i As Integer
            Dim dnglError As SmartDnglError
            Dim bytesReturned As Integer
            Dim status As Boolean = False
            Dim buffer(CY7C637XX.SECRETS_KEY_LEN) As Byte


            ' If signed read/write is supported
            If (dongle.options And CY7C637XX.QTS_OPTION_EXTENDED_SUPPORT) > 0 Then

                dongle.signature = CY7C637XX.SignatureBytesFromKey(reply)

                ' Use 32 bit signature if using >=2.3.x drivers
                If (((dongle.major = 2) And (dongle.minor >= 3)) Or (dongle.major >= 3)) Then

                    buffer = BitConverter.GetBytes(reply)

                    ' VB .NET does not allow a 0x00000000FFFFFFFF mask
                    ' reply &= 0x00000000FFFFFFFF
                    For i = (CY7C637XX.SECRETS_KEY_LEN - 1) To _
                       (CY7C637XX.SECRETS_KEY_LEN - 4) Step -1

                       buffer(i) = 0
                    Next i

                Else

                    ' Strict access requires SmartDongle
                    ' driver version 2.3.0 or greater.
                    If (dongle.options And CY7C637XX.QTS_OPTION_STRICT_ONLY) > 0 Then
                       Return SmartDnglError.UskErrRequiresNewerDriver
                    End If

                    reply = reply And &HFFFFFFFFFFFFF
                    buffer = BitConverter.GetBytes(reply)

                End If
            Else
                buffer = BitConverter.GetBytes(reply)
            End If

            ' Wait for task_loop processing to finish
            dnglError = waitForBusy(dongle.device, CY7C637XX.QTSBUSYWAIT)
            If dnglError <> SmartDnglError.UskOk Then
                Return dnglError
            End If

            ' Driver does not queue requests, retry on fail
            For i = 0 To CY7C637XX.QTS_IOCTL_RETRYS - 1
                status = Win32Methods.DeviceIoControl( _
                       dongle.device, _
                       IOCTL_QTSDONGLE_SEND_KEY(), _
                       buffer, _
                       CY7C637XX.SECRETS_KEY_LEN, _
                       IntPtr.Zero, _
                       0, _
                       bytesReturned, _
                       IntPtr.Zero)

                If status Then
                    Exit For
                End If

                Thread.Sleep(CY7C637XX.QTS_IOCTL_FAIL_WAIT)
            Next i

            If status = False Then
                Return SmartDnglError.UskErrIoctlSendKey
            End If

            Return SmartDnglError.UskOk
        End Function


        '
        ' Get the SmartDongle's Challenge.
        '
        Public Shared Function getKey( _
            ByVal dongle As Device, _
            ByRef challenge As Int64) As SmartDnglError

            Dim i As Integer
            Dim dnglError As SmartDnglError = SmartDnglError.UskOk
            Dim bytesReturned As Integer = 0
            Dim status As Boolean = False
            Dim b1 As IntPtr = Marshal.AllocHGlobal(CY7C637XX.SECRETS_KEY_LEN)
            Dim b2(CY7C637XX.SECRETS_KEY_LEN) As Byte


            ' Wait for task_loop processing to finish
            dnglError = waitForBusy(dongle.device, CY7C637XX.QTSBUSYWAIT)
            If dnglError <> SmartDnglError.UskOk Then
                challenge = 0
                Return dnglError
            End If

            ' Driver does not queue requests, retry on fail
            For i = 0 To CY7C637XX.QTS_IOCTL_RETRYS - 1
                status = Win32Methods.DeviceIoControl( _
                     dongle.device, _
                     IOCTL_QTSDONGLE_GET_KEY(), _
                     IntPtr.Zero, _
                     0, _
                     b1, _
                     CY7C637XX.SECRETS_KEY_LEN, _
                     bytesReturned, _
                     IntPtr.Zero)

                If status Then
                    Exit For
                End If
                Thread.Sleep(CY7C637XX.QTS_IOCTL_FAIL_WAIT)
            Next i

            If status = False Or bytesReturned < CY7C637XX.SECRETS_KEY_LEN Then
                challenge = 0
                dnglError = SmartDnglError.UskErrIoctlGetKey
            Else

                ' Must marshal bytes into byte array since C# expects a
                ' value in little endian form (if marshaling from IntPtr).
                For i = 0 To CY7C637XX.SECRETS_KEY_LEN - 1
                    b2(i) = Marshal.ReadByte(b1, i)
                Next i

                challenge = BitConverter.ToInt64(b2, 0)
            End If

            Marshal.FreeHGlobal(b1)
            Return dnglError
        End Function

        '
        ' Write to Port 0-3, must be set by firmware as an OUT port.
        '
        Public Shared Function writePort( _
            ByVal device As IntPtr, _
             ByVal port As Byte, _
             ByVal color As Byte) As SmartDnglError

            Dim i As Integer
            Dim bytesReturned As Integer
            Dim status As Boolean = False
            Dim in_buffer(2) As Byte


            in_buffer(0) = port
            in_buffer(1) = color

            For i = 0 To CY7C637XX.QTS_IOCTL_RETRYS - 1
                status = Win32Methods.DeviceIoControl( _
                     device, _
                     IOCTL_QTSDONGLE_WRITE_PORT(), _
                     in_buffer, _
                     2, _
                     IntPtr.Zero, _
                     0, _
                     bytesReturned, _
                     IntPtr.Zero)

                If status Then
                    Exit For
                End If

                Thread.Sleep(CY7C637XX.QTS_IOCTL_FAIL_WAIT)
            Next i

            If status = False Then
                Return SmartDnglError.UskErrIoctlQtsDongleSendPort
            End If

            Return SmartDnglError.UskOk
        End Function

        '
        ' Get the SmartDongle's security state: 0) Disabled
        '                                       1) Enabled
        Public Shared Function readSecurityState( _
            ByVal device As IntPtr, _
            ByRef state As Byte) As SmartDnglError

            Dim i As Integer
            Dim bytesReturned As Integer
            Dim status As Boolean = False
            Dim out_buffer(CY7C637XX.QTSPACKETSIZE) As Byte


            For i = 0 To CY7C637XX.QTS_IOCTL_RETRYS - 1
                status = Win32Methods.DeviceIoControl( _
                     device, _
                     IOCTL_QTSDONGLE_GET_SECURITY_STATE(), _
                     IntPtr.Zero, _
                     0, _
                     out_buffer, _
                     CY7C637XX.QTSPACKETSIZE, _
                     bytesReturned, _
                     IntPtr.Zero)

                If status Then
                    Exit For
                End If

                Thread.Sleep(CY7C637XX.QTS_IOCTL_FAIL_WAIT)
            Next i

            ' If status is true, state will be in first byte
            state = out_buffer(0)

            If status = False Then
                Return SmartDnglError.UskErrIoctlGetSecurityState
            End If

            Return SmartDnglError.UskOk
        End Function

        '
        ' Reset the SmartDongle's security state to 0  (Disabled)
        '
        Public Shared Function resetSecurityState(ByVal device As IntPtr) As SmartDnglError

            Dim i As Integer
            Dim bytesReturned As Integer = 0
            Dim status As Boolean = True


            For i = 0 To CY7C637XX.QTS_IOCTL_RETRYS - 1
                status = Win32Methods.DeviceIoControl( _
                     device, _
                     IOCTL_QTSDONGLE_RESET_SECURITY_STATE(), _
                     IntPtr.Zero, _
                     0, _
                     IntPtr.Zero, _
                     0, _
                     bytesReturned, _
                     IntPtr.Zero)

                If status Then
                    Exit For
                End If

                Thread.Sleep(CY7C637XX.QTS_IOCTL_FAIL_WAIT)
            Next i

            If status = False Then
                Return SmartDnglError.UskErrIoctlResetSecurityState
            End If

            Return SmartDnglError.UskOk
        End Function

        '
        '  Reset the SmartDongle's key state to 0 (QTSINITIAL)
        '
        Public Shared Function resetKeyState(ByVal device As IntPtr) As SmartDnglError

            Dim i As Integer
            Dim dnglError As SmartDnglError
            Dim bytesReturned As Integer = 0
            Dim status As Boolean = True


            ' Wait for task_loop processing to finish
            dnglError = waitForBusy(device, CY7C637XX.QTSBUSYWAIT)
            If dnglError <> SmartDnglError.UskOk Then
                Return dnglError
            End If

            For i = 0 To CY7C637XX.QTS_IOCTL_RETRYS - 1
                status = Win32Methods.DeviceIoControl( _
                     device, _
                     IOCTL_QTSDONGLE_RESET_KEY_STATE(), _
                     IntPtr.Zero, _
                     0, _
                     IntPtr.Zero, _
                     0, _
                     bytesReturned, _
                     IntPtr.Zero)

                If status Then
                    Exit For
                End If

                Thread.Sleep(CY7C637XX.QTS_IOCTL_FAIL_WAIT)
            Next i

            If status = False Then
                Return SmartDnglError.UskErrIoctlResetKeyState
            End If

            Return SmartDnglError.UskOk
        End Function

        '
        ' Request the SmartDongle to create a Challenge.
        '
        Public Shared Function requestHostKey(ByVal device As IntPtr) As SmartDnglError

            Dim i As Integer
            Dim dnglError As SmartDnglError
            Dim bytesReturned As Integer = 0
            Dim status As Boolean = True


            ' Wait for task_loop processing to finish
            dnglError = waitForBusy(device, CY7C637XX.QTSBUSYWAIT)
            If dnglError <> SmartDnglError.UskOk Then
                Return dnglError
            End If

            For i = 0 To CY7C637XX.QTS_IOCTL_RETRYS - 1
                status = Win32Methods.DeviceIoControl( _
                     device, _
                     IOCTL_QTSDONGLE_SEND_HOST_VERIFY_REQUEST(), _
                     IntPtr.Zero, _
                     0, _
                     IntPtr.Zero, _
                     0, _
                     bytesReturned, _
                     IntPtr.Zero)

                If status Then
                    Exit For
                End If

                Thread.Sleep(CY7C637XX.QTS_IOCTL_FAIL_WAIT)
            Next i

            If status = False Then
                Return SmartDnglError.UskErrIoctlSendHostVerifyRequest
            End If

            Return SmartDnglError.UskOk
        End Function

        '
        ' Send a page of data out to the SmartDongle's write buffer, use
        ' readPageFlags to poll the status of the EEPROM write buffer.
        Public Shared Function writeEpromPage( _
             ByVal address As Int16, _
             ByVal page() As Byte, _
             ByVal dongle As Device) As SmartDnglError

            Dim i, j As Integer
            Dim bytesReturned As Integer = 0
            Dim status As Boolean = True
            Dim in_size As Integer
            Dim address_bytes(CY7C637XX.ADDR_LENGTH) As Byte
            Dim buffer(2 * CY7C637XX.EEPROMPAGESIZE) As Byte


            ' Int16 address to big endian byte array
            address_bytes = BitConverter.GetBytes(address)

            If (dongle.options And CY7C637XX.QTS_OPTION_EXTENDED_SUPPORT) > 0 Then

                ' 32 bit Signature
                j = 0
                For i = 0 To CY7C637XX.SIG_LENGTH - 1
                    buffer(j) = dongle.signature(i)
                    j += 1
                Next i

                ' EEPROM address
                For i = 0 To CY7C637XX.ADDR_LENGTH - 1
                    buffer(j) = address_bytes(i)
                    j += 1
                Next i

                ' 16 bytes of page data
                For i = 0 To CY7C637XX.EEPROMPAGESIZE - 1
                    buffer(j) = page(i)
                    j += 1
                Next i

                in_size = CY7C637XX.SIG_LENGTH + _
                     CY7C637XX.ADDR_LENGTH + _
                     CY7C637XX.EEPROMPAGESIZE
            Else

                ' EEPROM address
                j = 0
                For i = 0 To CY7C637XX.ADDR_LENGTH - 1
                    buffer(j) = address_bytes(i)
                    j += 1
                Next i

                ' 16 bytes of page data
                For i = 0 To CY7C637XX.EEPROMPAGESIZE - 1
                    buffer(j) = page(i)
                    j += 1
                Next i

                in_size = CY7C637XX.ADDR_LENGTH + CY7C637XX.EEPROMPAGESIZE
            End If

            For i = 0 To CY7C637XX.QTS_IOCTL_RETRYS - 1
                status = Win32Methods.DeviceIoControl( _
                     dongle.device, _
                     IOCTL_QTSDONGLE_WRITE_EEPROM(), _
                     buffer, _
                     in_size, _
                     IntPtr.Zero, _
                     0, _
                     bytesReturned, _
                     IntPtr.Zero)

                If status Then
                    Exit For
                End If

                Thread.Sleep(CY7C637XX.QTS_IOCTL_FAIL_WAIT)
            Next i

            If status = False Then
                Return SmartDnglError.UskErrIoctlWriteEeprom
            End If

            Return SmartDnglError.UskOk
        End Function

        '
        '  Send the address of a page in EEPROM that will be read
        '   on the next ReadEpromPrefetch.
        Public Shared Function requestEpromPrefetch( _
            ByVal device As IntPtr, _
            ByVal address As Int16) As SmartDnglError

            Dim i As Integer
            Dim bytesReturned As Integer = 0
            Dim address_bytes(CY7C637XX.ADDR_LENGTH) As Byte
            Dim status As Boolean = True


            ' Int16 address to big endian byte array
            address_bytes = BitConverter.GetBytes(address)

            ' Driver does not queue requests, retry on fail
            For i = 0 To CY7C637XX.QTS_IOCTL_RETRYS - 1
                status = Win32Methods.DeviceIoControl( _
                     device, _
                     IOCTL_QTSDONGLE_REQUEST_PAGE(), _
                     address_bytes, CY7C637XX.ADDR_LENGTH, _
                     IntPtr.Zero, _
                     0, bytesReturned, _
                     IntPtr.Zero)

                If status Then
                    Exit For
                End If

                Thread.Sleep(CY7C637XX.QTS_IOCTL_FAIL_WAIT)
            Next i

            If status = False Then
                Return SmartDnglError.UskErrIoctlRequestPage
            End If

            Return SmartDnglError.UskOk
        End Function

        '
        ' Get the data currently in the SmartDongle's EEPROM page read buffer.
        ' Use readPageFlags to poll the status of the EEPROM read/write buffer.
        Public Shared Function readEpromPrefetch( _
             ByRef buffer() As Byte, _
             ByRef size As Integer, _
             ByVal dongle As Device) As SmartDnglError

            Dim i As Integer
            Dim bytesReturned As Integer = 0
            Dim status As Boolean = True
            Dim dnglError As SmartDnglError


            dnglError = IOCTL_Command.waitForBusy( _
                dongle.device, CY7C637XX.QTSBUSYWAIT)
            If dnglError <> SmartDnglError.UskOk Then
                Return dnglError
            End If

            '  If the SmartDongle supports "signed" read/write packets.
            If (dongle.options And CY7C637XX.QTS_OPTION_EXTENDED_SUPPORT) > 0 Then

                '  If using the 2.3.0 driver, then send the 32 bit
                '  "signature" in the output buffer.
                If dongle.major = 2 And dongle.minor >= 3 Or dongle.major >= 3 Then

                    Dim in_buffer(CY7C637XX.QTSPACKETSIZE) As Byte

                    ' 32 bit Signature
                    in_buffer(0) = dongle.signature(2)
                    in_buffer(1) = dongle.signature(3)
                    in_buffer(2) = dongle.signature(0)
                    in_buffer(3) = dongle.signature(1)

                    For i = 0 To CY7C637XX.QTS_IOCTL_RETRYS - 1
                       status = Win32Methods.DeviceIoControl( _
                           dongle.device, _
                           IOCTL_QTSDONGLE_READ_EEPROM(), _
                           in_buffer, _
                           CY7C637XX.QTSPACKETSIZE, _
                           buffer, _
                           8 * CY7C637XX.EEPROMPAGESIZE, _
                           bytesReturned, _
                           IntPtr.Zero)

                       If status Then
                          Exit For
                       End If

                       Thread.Sleep(CY7C637XX.QTS_IOCTL_FAIL_WAIT)
                    Next i

                    ' 12 bit signature
                Else

                    'Dim out_size As Integer = CInt(BitConverter.ToInt16(sig, 0))
                    Dim out_size As Integer = _
                       CInt(BitConverter.ToInt16(dongle.signature, 0))
                    out_size = CInt((out_size And &HFFF0) / 16) ' >>4

                    ' A key value below the size of the data to be read
                    ' will return an dnglError from the bus driver.
                    If out_size < (CY7C637XX.CP2_PAGE_PREFETCH_ENABLED * CY7C637XX.EEPROMPAGESIZE) Then
                       out_size = out_size Or &H100
                    End If

                    For i = 0 To CY7C637XX.QTS_IOCTL_RETRYS - 1
                       status = Win32Methods.DeviceIoControl( _
                           dongle.device, _
                           IOCTL_QTSDONGLE_READ_EEPROM(), _
                           IntPtr.Zero, _
                           0, buffer, _
                           out_size, _
                           bytesReturned, _
                           IntPtr.Zero)

                       If status Then
                          Exit For
                       End If

                       Thread.Sleep(CY7C637XX.QTS_IOCTL_FAIL_WAIT)
                    Next i
                End If
            Else
                For i = 0 To CY7C637XX.QTS_IOCTL_RETRYS - 1
                    status = Win32Methods.DeviceIoControl( _
                        dongle.device, _
                        IOCTL_QTSDONGLE_READ_EEPROM(), _
                        IntPtr.Zero, _
                        0, _
                        buffer, _
                        8 * CY7C637XX.EEPROMPAGESIZE, _
                        bytesReturned, _
                        IntPtr.Zero)

                    If status Then
                       Exit For
                    End If

                    Thread.Sleep(CY7C637XX.QTS_IOCTL_FAIL_WAIT)
                Next i
            End If

            If status = False Then
                size = 0
                Return SmartDnglError.UskErrIoctlReadEeprom
            End If

            size = bytesReturned
            Return SmartDnglError.UskOk
        End Function


        '
        ' Reboot SmartDongle, ignore return dnglError.
        '
        Public Shared Function SoftReset(ByVal device As IntPtr) As SmartDnglError

            Dim bytesReturned As Integer


            Win32Methods.DeviceIoControl( _
                 device, _
                 IOCTL_QTSDONGLE_SOFT_RESET(), _
                 IntPtr.Zero, _
                 0, _
                 IntPtr.Zero, _
                 0, _
                 bytesReturned, _
                 IntPtr.Zero)

            Return SmartDnglError.UskOk
        End Function

        '
        ' For SmartDongle >=v1.50 using >= v2.3 driver only.
        ' This will configure the SmartDongle for Strict Access.
        ' A new key and 32 bit signatures are required for each user
        ' read/write transaction against user memory.
        '
        ' Caution: This configuration is permanent.
        '          SmartDongle must be sent back to the manufacturer to
        '          be reset to use the older drivers (<= v2.2.0).
        Public Shared Function setStrictOnly(ByVal options As Byte) As SmartDnglError

            Dim i As Integer
            Dim dnglError As SmartDnglError
            Dim bytesReturned As Integer
            Dim status As Boolean = False
            Dim dongle As New Device


            '  Keep customer from burning memory with repeated requests.
            If (options And CY7C637XX.QTS_OPTION_STRICT_ONLY) > 0 Then
                Return SmartDnglError.UskOk
            End If

            ' Requires >=2.3.x drivers
            If (dongle.major = 2 And dongle.minor >= 3 Or dongle.major >= 3) = False Then
                Return SmartDnglError.UskErrRequiresNewerDriver
            End If

            ' Wait for task_loop processing to finish
            dnglError = waitForBusy(dongle.device, CY7C637XX.QTSBUSYWAIT)
            If dnglError <> SmartDnglError.UskOk Then
                Return dnglError
            End If

            For i = 0 To CY7C637XX.QTS_IOCTL_RETRYS - 1
                status = Win32Methods.DeviceIoControl( _
                     dongle.device, _
                     IOCTL_QTSDONGLE_SET_STRICT(), _
                     IntPtr.Zero, _
                     0, _
                     IntPtr.Zero, _
                     0, _
                     bytesReturned, _
                     IntPtr.Zero)

                If status Then
                    Exit For
                End If

                Thread.Sleep(CY7C637XX.QTS_IOCTL_FAIL_WAIT)
            Next i

            If status = False Then
                Return SmartDnglError.UskErrIoctlSetStrict
            End If

            ' Update SmartDongle options
            options = options Or CY7C637XX.QTS_OPTION_STRICT_ONLY

            Return SmartDnglError.UskOk
        End Function


        '
        ' Read SmartDongle's current eprom bits.
        '
        ' This is the encryption for the data on the SmartDongle memory
        ' that protects the memory from a physical attack only.
        '
        '
        '  Modes:   0x00 QTSSTRONGCRYPT    AES Encryption
        '           0x01 QTSCRYPTDISABLED  No Encryption
        '           0x02 QTSFASTCRYPT      Fast, lightwieght encryption
        Public Shared Function GetMode( _
            ByVal device As IntPtr, _
            ByRef mode As Byte) As SmartDnglError

            Dim i As Integer
            Dim dnglError As SmartDnglError
            Dim bytesReturned As Integer
            Dim status As Boolean = False
            Dim out_buffer(CY7C637XX.QTSPACKETSIZE) As Byte


            ' Wait for task_loop processing to finish
            dnglError = waitForBusy(device, CY7C637XX.QTSBUSYWAIT)
            If dnglError <> SmartDnglError.UskOk Then
                mode = &H0
                Return dnglError
            End If

            For i = 0 To CY7C637XX.QTS_IOCTL_RETRYS - 1
                status = Win32Methods.DeviceIoControl( _
                     device, _
                     IOCTL_QTSDONGLE_GET_MODE(), _
                     IntPtr.Zero, _
                     0, _
                     out_buffer, _
                     CY7C637XX.QTSPACKETSIZE, _
                     bytesReturned, _
                     IntPtr.Zero)

                If status Then
                    Exit For
                End If

                Thread.Sleep(CY7C637XX.QTS_IOCTL_FAIL_WAIT)
            Next i

            ' If status is true, state will be in first byte.
            mode = out_buffer(0)

            ' Mask out extraneous bits
            mode = _
                CByte( _
                    mode And _
                    (UserOption.QTS_OPTION_ON_CHIP_EPROM_PROTECT_DISABLE Or _
                    UserOption.QTS_OPTION_ON_CHIP_EPROM_PROTECT_FAST))

            If status = False Then
                Return SmartDnglError.UskErrIoctlGetMode
            End If

            Return SmartDnglError.UskOk
        End Function


        '
        ' Set SmartDongle eprom mode bits.
        '
        ' This is the encryption for the data on the SmartDongle
        ' that protects the EEPROM memory from a physical attack.
        '
        '  Modes:   0x00 QTSSTRONGCRYPT    AES Encryption
        '           0x01 QTSCRYPTDISABLED  No Encryption
        '           0x02 QTSFASTCRYPT      Fast, lightweight encryption.
        Public Shared Function SetMode( _
            ByVal device As IntPtr, _
            ByVal mode As Byte) As SmartDnglError

            Dim i As Integer
            Dim dnglError As SmartDnglError
            Dim bytesReturned As Integer
            Dim status As Boolean = False
            Dim in_buffer(CY7C637XX.QTSPACKETSIZE) As Byte


            ' Wait for task_loop processing to finish
            dnglError = waitForBusy(device, CY7C637XX.QTSBUSYWAIT)
            If dnglError <> SmartDnglError.UskOk Then
                Return dnglError
            End If

            in_buffer(0) = mode

            For i = 0 To CY7C637XX.QTS_IOCTL_RETRYS - 1
                status = Win32Methods.DeviceIoControl( _
                     device, _
                     IOCTL_QTSDONGLE_SET_MODE(), _
                     in_buffer, CY7C637XX.QTSPACKETSIZE, _
                     IntPtr.Zero, _
                     0, _
                     bytesReturned, _
                     IntPtr.Zero)

                If status Then
                    Exit For
                End If

                Thread.Sleep(CY7C637XX.QTS_IOCTL_FAIL_WAIT)
            Next i

            If status = False Then
                Return SmartDnglError.UskErrIoctlSetMode
            End If

            Return SmartDnglError.UskOk
        End Function


        '
        ' Read SmartDongle's extended functionality bits are mapped with
        ' eprom error bits.
        Public Shared Function GetOpt(ByRef dongle As Device) As SmartDnglError

            Dim i As Integer
            Dim dnglError As SmartDnglError
            Dim bytesReturned As Integer
            Dim status As Boolean = False
            Dim out_buffer(CY7C637XX.QTSPACKETSIZE) As Byte
            Dim fixedOptions As Byte


            ' clear options
            dongle.options = 0

            ' Wait for task_loop processing to finish.
            dnglError = waitForBusy(dongle.device, CY7C637XX.QTSBUSYWAIT)
            If dnglError <> SmartDnglError.UskOk Then
                Return dnglError
            End If

            For i = 0 To CY7C637XX.QTS_IOCTL_RETRYS - 1
                status = Win32Methods.DeviceIoControl( _
                     dongle.device, _
                     IOCTL_QTSDONGLE_GET_EEPROM_ERROR_FLAG(), _
                     IntPtr.Zero, _
                     0, _
                     out_buffer, _
                     CY7C637XX.QTSPACKETSIZE, _
                     bytesReturned, _
                     IntPtr.Zero)

                If status Then
                    Exit For
                End If

                Thread.Sleep(CY7C637XX.QTS_IOCTL_FAIL_WAIT)
            Next i

            ' If status is true, state will be in first byte
            fixedOptions = out_buffer(0)

            ' Map firmware flags to options flags
            If (fixedOptions And CY7C637XX.QTS_FIRMWARE_EXTENDED_SUPPORT2) > 0 Then
                dongle.options = _
                    dongle.options Or CY7C637XX.QTS_OPTION_EXTENDED_SUPPORT2
            End If

            If (fixedOptions And CY7C637XX.QTS_FIRMWARE_EXTENDED_SUPPORT) > 0 Then
                dongle.options = _
                    dongle.options Or CY7C637XX.QTS_OPTION_EXTENDED_SUPPORT
            End If

            If (fixedOptions And CY7C637XX.QTS_FIRMWARE_STRICT_ONLY) > 0 Then
                dongle.options = _
                    dongle.options Or CY7C637XX.QTS_OPTION_STRICT_ONLY
            End If

            If status = False Then
                Return SmartDnglError.UskErrIoctlGetEepromErrorFlag
            End If

            Return SmartDnglError.UskOk
        End Function


        '
        '  Some functionality only supported in new, unsigned driver.
        '
        '  out_buffer[0] =  QTSDRIVER_VERSION & 0x000000FF;        // m3
        '  out_buffer[1] = (QTSDRIVER_VERSION & 0x0000FF00) >> 8;  // m2
        '  out_buffer[2] = (QTSDRIVER_VERSION & 0x00FF0000) >> 16; // minor
        '  out_buffer[3] = (QTSDRIVER_VERSION & 0xFF000000) >> 24; // major
        Public Shared Function GetDriverVersion(ByRef dongle As Device) As SmartDnglError

            Dim i As Integer
            Dim bytesReturned As Integer
            Dim status As Boolean = True
            Dim out_buffer(8 * CY7C637XX.EEPROMPAGESIZE) As Byte


            ' Driver does not queue requests, retry on fail
            For i = 0 To CY7C637XX.QTS_IOCTL_RETRYS - 1
                status = Win32Methods.DeviceIoControl( _
                     dongle.device, _
                     IOCTL_QTSDONGLE_GET_DRIVER_REVISION(), _
                     IntPtr.Zero, _
                     0, _
                     out_buffer, _
                     8 * CY7C637XX.EEPROMPAGESIZE, _
                     bytesReturned, _
                     IntPtr.Zero)

                If status Then
                    Exit For
                End If

                Thread.Sleep(CY7C637XX.QTS_IOCTL_FAIL_WAIT)
            Next i

            If status = False Then
                Return SmartDnglError.UskErrIoctlGetDriverRevision
            End If

            ' Copy over version byte fields.
            dongle.minor = out_buffer(2)
            dongle.major = out_buffer(3)

            Return SmartDnglError.UskOk
        End Function


        '
        '   Get SmartDongle serial number.
        '
        Public Shared Function getSerialNumber( _
             ByVal device As IntPtr, ByRef sn() As Char) As SmartDnglError

            Dim i As Integer
            Dim dnglError As SmartDnglError
            Dim bytesReturned As Integer
            Dim status As Boolean = True
            Dim in_buffer(CY7C637XX.QTSPACKETSIZE) As Byte
            Dim out_buffer(8 * CY7C637XX.EEPROMPAGESIZE) As Byte
            sn = New Char(CP2_SERIALNUM_LENGTH) {}


            ' Wait for task_loop processing to finish
            dnglError = waitForBusy(device, CY7C637XX.QTSBUSYWAIT)
            If dnglError <> SmartDnglError.UskOk Then
                Return dnglError
            End If

            in_buffer(0) = &H3 ' Index for serial number fixed at 0x03

            For i = 0 To CY7C637XX.QTS_IOCTL_RETRYS - 1
                status = Win32Methods.DeviceIoControl( _
                     device, _
                     IOCTL_QTSDONGLE_GET_STRING_DESCRIPTOR(), _
                     in_buffer, _
                     CY7C637XX.QTSPACKETSIZE, _
                     out_buffer, _
                     8 * CY7C637XX.EEPROMPAGESIZE, _
                     bytesReturned, IntPtr.Zero)

                If status Then
                    Exit For
                End If

                Thread.Sleep(CY7C637XX.QTS_IOCTL_FAIL_WAIT)
            Next i

            If status = False Then
                Return SmartDnglError.UskErrIoctlGetStringDescriptor
            End If

            ' The length member of the returned USB descriptor string
            ' must be this length.
            If out_buffer(0) <> CY7C637XX.CP2_SER_NUMB_STRING_LENGTH Then
                Return SmartDnglError.UskErrDataFormat
            End If

            For i = 0 To CP2_SERIALNUM_LENGTH - 1
                sn(i) = Convert.ToChar(out_buffer(2 + i * 2))
            Next i

            Return SmartDnglError.UskOk
        End Function


        '
        ' Wait for EPROM access status byte to indicate that requested
        ' EPROM pages are ready or an error has occurred.
        Public Shared Function waitForPageRead( _
            ByVal device As IntPtr, _
            ByVal period As Integer, _
            ByRef flags As Byte) As SmartDnglError

            Dim data As New Byte
            Dim dnglError As SmartDnglError
            Dim trys As Integer = 10
            Dim i As Integer


            If period <= 0 Then
                flags = 0
                Return SmartDnglError.UskErrUnknown
            End If

            For i = 0 To trys - 1
                dnglError = readPageFlags(device, data)
                If dnglError <> SmartDnglError.UskOk Then
                    flags = 0
                    Return dnglError
                End If

                ' Any read-related bit set mean ready
                If (data And CY7C637XX.QTS_READ_PAGE_READY_MASK) > 0 Then
                    flags = data
                    Return SmartDnglError.UskOk
                End If

                Thread.Sleep(CInt(System.Math.Floor(period / trys)))
            Next i

            flags = 0

            Return SmartDnglError.UskErrEepromReadTimeout
        End Function


        '   Get SmartDongle's EPROM access status byte.
        Public Shared Function readPageFlags( _
             ByVal device As IntPtr, ByRef flags As Byte) As SmartDnglError

            Dim i As Integer
            Dim bytesReturned As Integer
            Dim status As Boolean = False
            Dim out_buffer(CY7C637XX.QTSPACKETSIZE) As Byte


            For i = 0 To CY7C637XX.QTS_IOCTL_RETRYS - 1
                status = Win32Methods.DeviceIoControl( _
                     device, _
                     IOCTL_QTSDONGLE_GET_PAGE_FLAGS(), _
                     IntPtr.Zero, _
                     0, _
                     out_buffer, _
                     CY7C637XX.QTSPACKETSIZE, _
                     bytesReturned, _
                     IntPtr.Zero)

                If status Then
                    Exit For
                End If

                Thread.Sleep(CY7C637XX.QTS_IOCTL_FAIL_WAIT)
            Next i

            ' If status is true, state flags will be in first byte
            flags = out_buffer(0)

            If status = False Then
                Return SmartDnglError.UskErrIoctlGetPageFlags
            End If

            Return SmartDnglError.UskOk
        End Function


        '
        ' Wait for a particular challenge/request state.
        '
        ' Note: Removed the host challenge as in v1.41, so states are:
        '
        '    initial state:  0
        '    challege state: v1.41:0x02  v1.42:0x01
        Public Shared Function waitForKeyState( _
            ByVal device As IntPtr, _
            ByVal state As Byte, _
            ByVal period As Integer) As SmartDnglError

            Dim data As New Byte
            Dim key_state, required_state As Boolean
            Dim dnglError As SmartDnglError
            Dim trys As Integer = 10
            Dim i As Integer
            Const USK_KEY_STATE_MASK As Byte = &H3


            ' boolean
            required_state = (state And USK_KEY_STATE_MASK) > 0

            If period <= 0 Then
                Return SmartDnglError.UskErrUnknown
            End If

            For i = 0 To trys - 1
                dnglError = readKeyState(device, data)
                If dnglError <> SmartDnglError.UskOk Then
                    Return dnglError
                End If

                ' boolean
                key_state = (data And USK_KEY_STATE_MASK) > 0

                If required_state = key_state Then
                    Return SmartDnglError.UskOk
                End If

                Thread.Sleep(CInt(System.Math.Floor(period / trys)))
            Next i

            Return SmartDnglError.UskErrKeyStateTimeout
        End Function

        Shared Function readKeyState( _
            ByVal device As IntPtr, _
            ByRef state As Byte) As SmartDnglError
            Dim i As Integer
            Dim bytesReturned As Integer
            Dim status As Boolean = False
            Dim out_buffer(CY7C637XX.QTSPACKETSIZE) As Byte


            For i = 0 To CY7C637XX.QTS_IOCTL_RETRYS - 1
                status = Win32Methods.DeviceIoControl( _
                     device, _
                     IOCTL_QTSDONGLE_GET_KEY_STATE(), _
                     IntPtr.Zero, _
                     0, _
                     out_buffer, _
                     CY7C637XX.QTSPACKETSIZE, _
                     bytesReturned, _
                     IntPtr.Zero)

                If status Then
                    Exit For
                End If

                Thread.Sleep(CY7C637XX.QTS_IOCTL_FAIL_WAIT)
            Next i

            ' If status is true, state will be in first byte
            state = out_buffer(0)

            If status = False Then
                Return SmartDnglError.UskErrIoctlGetKeyState
            End If

            Return SmartDnglError.UskOk
        End Function


        '
        '   Wait Time Write Cycle flag to reset.
        Public Shared Function waitForTwc( _
            ByVal device As IntPtr, _
             ByVal period As Integer) As SmartDnglError

            Dim data As New Byte
            Dim dnglError As SmartDnglError
            Dim trys As Integer = 10
            Dim i As Integer

            If period <= 0 Then
                Return SmartDnglError.UskErrUnknown
            End If

            For i = 0 To trys - 1
                dnglError = readTwcFlag(device, data)
                If dnglError <> SmartDnglError.UskOk Then
                    Return dnglError
                End If

                If data = CY7C637XX.QTSREADY Then
                    Return SmartDnglError.UskOk
                End If

                Thread.Sleep(CInt(System.Math.Floor(period / trys)))
            Next i

            Return SmartDnglError.UskErrTwcResetTimeout
        End Function


        Shared Function readTwcFlag( _
            ByVal device As IntPtr, ByRef state As Byte) As SmartDnglError

            Dim i As Integer
            Dim bytesReturned As Integer
            Dim status As Boolean = False
            Dim out_buffer(CY7C637XX.QTSPACKETSIZE) As Byte


            For i = 0 To CY7C637XX.QTS_IOCTL_RETRYS - 1
                status = Win32Methods.DeviceIoControl( _
                     device, _
                     IOCTL_QTSDONGLE_GET_TWC_FLAG(), _
                     IntPtr.Zero, _
                     0, out_buffer, _
                     CY7C637XX.QTSPACKETSIZE, _
                     bytesReturned, _
                     IntPtr.Zero)

                If status Then
                    Exit For
                End If

                Thread.Sleep(CY7C637XX.QTS_IOCTL_FAIL_WAIT)
            Next i

            ' If status is true, state will be in first byte
            state = out_buffer(0)

            If status = False Then
                Return SmartDnglError.UskErrIoctlGetTwcFlag
            End If

            Return SmartDnglError.UskOk
        End Function 'readTwcFlag


        '
        ' Wait for busy flag to be reset
        ' backwards compatible with <=1.40 Firmware
        Public Shared Function waitForBusy( _
            ByVal device As IntPtr, _
             ByVal period As Integer) As SmartDnglError

            Dim data As New Byte
            Dim dnglError As SmartDnglError
            Dim trys As Integer = 10
            Dim i As Integer


            If period <= 0 Then
                Return SmartDnglError.UskErrUnknown
            End If

            For i = 0 To trys - 1
                dnglError = readBusyFlag(device, data)
                If dnglError <> SmartDnglError.UskOk Then
                    Return dnglError
                End If

                If data = CY7C637XX.QTSREADY Then
                    Return SmartDnglError.UskOk
                End If

                Thread.Sleep(CInt(System.Math.Floor(period / trys)))
            Next i

            Return SmartDnglError.UskErrBusyTimeout
        End Function


        Shared Function readBusyFlag( _
            ByVal device As IntPtr, _
            ByRef state As Byte) As SmartDnglError

            Dim i As Integer
            Dim bytesReturned As Integer
            Dim status As Boolean = False
            Dim out_buffer(CY7C637XX.QTSPACKETSIZE) As Byte


            For i = 0 To CY7C637XX.QTS_IOCTL_RETRYS - 1
                status = Win32Methods.DeviceIoControl( _
                     device, _
                     IOCTL_QTSDONGLE_GET_BUSY_FLAG(), _
                     IntPtr.Zero, _
                     0, _
                     out_buffer, _
                     CY7C637XX.QTSPACKETSIZE, _
                     bytesReturned, _
                     IntPtr.Zero)

                If status Then
                    Exit For
                End If

                Thread.Sleep(CY7C637XX.QTS_IOCTL_FAIL_WAIT)
            Next i

            ' If status is true, state will be in first byte
            state = out_buffer(0)

            If status = False Then
                Return SmartDnglError.UskErrIoctlGetBusyFlag
            End If

            Return SmartDnglError.UskOk
        End Function 'readBusyFlag



        ' Attempt an innocuous command.
        '
        ' Return with status within dwMilliseconds.
        ' Time argument for device mutex access, set to zero if
        ' simply querying for "is this SmartDongle present?".
        Public Shared Function isDeviceReady( _
            ByVal device As IntPtr, _
            ByVal dwMilliseconds As Integer) As SmartDnglError

            Dim i As Integer
            Dim dnglError As SmartDnglError
            Dim bytesReturned As Integer
            Dim status As Boolean = True
            Dim in_buffer(CY7C637XX.QTSPACKETSIZE) As Byte
            Dim out_buffer(8 * CY7C637XX.EEPROMPAGESIZE) As Byte


            If dwMilliseconds > 0 Then
                dnglError = waitForBusy(device, CY7C637XX.QTSBUSYWAIT)
                If dnglError <> SmartDnglError.UskOk Then
                    Return dnglError
                End If
            End If

            in_buffer(0) = &H3 ' Index for serial number fixed at 0x03

            For i = 0 To CY7C637XX.QTS_IOCTL_RETRYS - 1

                status = Win32Methods.DeviceIoControl( _
                     device, _
                     IOCTL_QTSDONGLE_GET_STRING_DESCRIPTOR(), _
                     in_buffer, _
                     CY7C637XX.QTSPACKETSIZE, _
                     out_buffer, _
                     8 * CY7C637XX.EEPROMPAGESIZE, _
                     bytesReturned, _
                     IntPtr.Zero)

                If status Then
                    Exit For
                End If

                Thread.Sleep(CY7C637XX.QTS_IOCTL_FAIL_WAIT)
            Next i

            If status = False Then
                Return SmartDnglError.UskErrNoDevice
            End If

            '
            ' The length member of the returned USB descriptor string must be this length.
            If out_buffer(0) <> CY7C637XX.CP2_SER_NUMB_STRING_LENGTH Or out_buffer(1) <> CY7C637XX.USB_STRING_DESCRIPTOR_TYPE Then
                Return SmartDnglError.UskErrNoDevice
            End If

            Return SmartDnglError.UskOk
        End Function 'isDeviceReady

        '   winioctl.h
        Const FILE_DEVICE_UNKNOWN As Integer = &H22
        Const METHOD_BUFFERED As Integer = 0
        Const FILE_ANY_ACCESS As Integer = 0


        Private Shared Function CTL_CODE( _
             ByVal DeviceType As Int32, _
             ByVal DeviceFunction As Int32, _
             ByVal Method As Int32, _
             ByVal Access As Int32) As Int32

            ' DeviceType<<16 | DeviceType<<14 | DeviceFunction<<4 | Method
            Return (DeviceType * &H10000) Or (Access * &H4000) Or (DeviceFunction * (4)) Or Method
        End Function

        ' sd_user.h
        Private Const QTSDONGLE_IOCTL_INDEX As Integer = &H800

        Private Shared Function _
             IOCTL_QTSDONGLE_GET_DEVICE_DESCRIPTOR() As Int32

            Return CTL_CODE( _
                 FILE_DEVICE_UNKNOWN, _
                 QTSDONGLE_IOCTL_INDEX + 1, _
                 METHOD_BUFFERED, FILE_ANY_ACCESS)
        End Function

        Private Shared Function _
             IOCTL_QTSDONGLE_GET_CONFIG_DESCRIPTOR() As Int32

            Return CTL_CODE( _
                 FILE_DEVICE_UNKNOWN, _
                 QTSDONGLE_IOCTL_INDEX + 2, _
                 METHOD_BUFFERED, _
                 FILE_ANY_ACCESS)
        End Function

        Private Shared Function _
             IOCTL_QTSDONGLE_RESET_PIPE() As Int32

            Return CTL_CODE( _
                 FILE_DEVICE_UNKNOWN, _
                 QTSDONGLE_IOCTL_INDEX + 6, _
                 METHOD_BUFFERED, _
                 FILE_ANY_ACCESS)
        End Function

        Private Shared Function _
             IOCTL_QTSDONGLE_GET_DRIVER_REVISION() As Int32

            Return CTL_CODE( _
                 FILE_DEVICE_UNKNOWN, _
                 QTSDONGLE_IOCTL_INDEX + 7, _
                 METHOD_BUFFERED, _
                 FILE_ANY_ACCESS)
        End Function

        Private Shared Function _
             IOCTL_QTSDONGLE_GET_STRING_DESCRIPTOR() As Int32

            Return CTL_CODE( _
                 FILE_DEVICE_UNKNOWN, _
                 QTSDONGLE_IOCTL_INDEX + 8, _
                 METHOD_BUFFERED, _
                 FILE_ANY_ACCESS)
        End Function

        Private Shared Function _
             IOCTL_QTSDONGLE_SEND_KEY() As Int32

            Return CTL_CODE( _
                 FILE_DEVICE_UNKNOWN, _
                 QTSDONGLE_IOCTL_INDEX + &HA, _
                 METHOD_BUFFERED, _
                 FILE_ANY_ACCESS)
        End Function

        Private Shared Function _
             IOCTL_QTSDONGLE_GET_KEY() As Int32

            Return CTL_CODE( _
                 FILE_DEVICE_UNKNOWN, _
                 QTSDONGLE_IOCTL_INDEX + &HB, _
                 METHOD_BUFFERED, _
                 FILE_ANY_ACCESS)
        End Function

        Private Shared Function _
             IOCTL_QTSDONGLE_WRITE_PORT() As Int32

            Return CTL_CODE( _
                 FILE_DEVICE_UNKNOWN, _
                 QTSDONGLE_IOCTL_INDEX + &HC, _
                 METHOD_BUFFERED, _
                 FILE_ANY_ACCESS)
        End Function

        Private Shared Function _
             IOCTL_QTSDONGLE_READ_EEPROM() As Int32

            Return CTL_CODE( _
                 FILE_DEVICE_UNKNOWN, _
                 QTSDONGLE_IOCTL_INDEX + &HE, _
                 METHOD_BUFFERED, _
                 FILE_ANY_ACCESS)
        End Function

        Private Shared Function _
             IOCTL_QTSDONGLE_WRITE_EEPROM() As Int32

            Return CTL_CODE( _
                 FILE_DEVICE_UNKNOWN, _
                 QTSDONGLE_IOCTL_INDEX + &HF, _
                 METHOD_BUFFERED, _
                 FILE_ANY_ACCESS)
        End Function

        Private Shared Function _
             IOCTL_QTSDONGLE_GET_SECURITY_STATE() As Int32

            Return CTL_CODE( _
                 FILE_DEVICE_UNKNOWN, _
                 QTSDONGLE_IOCTL_INDEX + &H11, _
                 METHOD_BUFFERED, _
                 FILE_ANY_ACCESS)
        End Function

        Private Shared Function _
             IOCTL_QTSDONGLE_RESET_SECURITY_STATE() As Int32

            Return CTL_CODE( _
                 FILE_DEVICE_UNKNOWN, _
                 QTSDONGLE_IOCTL_INDEX + &H12, _
                 METHOD_BUFFERED, FILE_ANY_ACCESS)
        End Function

        Private Shared Function _
             IOCTL_QTSDONGLE_GET_KEY_STATE() As Int32

            Return CTL_CODE( _
                 FILE_DEVICE_UNKNOWN, _
                 QTSDONGLE_IOCTL_INDEX + &H13, _
                 METHOD_BUFFERED, _
                 FILE_ANY_ACCESS)
        End Function

        Private Shared Function _
             IOCTL_QTSDONGLE_RESET_KEY_STATE() As Int32

            Return CTL_CODE( _
                 FILE_DEVICE_UNKNOWN, _
                 QTSDONGLE_IOCTL_INDEX + &H14, _
                 METHOD_BUFFERED, _
                 FILE_ANY_ACCESS)
        End Function

        Private Shared Function _
             IOCTL_QTSDONGLE_SEND_HOST_VERIFY_REQUEST() As Int32

            Return CTL_CODE( _
                 FILE_DEVICE_UNKNOWN, _
                 QTSDONGLE_IOCTL_INDEX + &H15, _
                 METHOD_BUFFERED, _
                 FILE_ANY_ACCESS)
        End Function

        Private Shared Function _
             IOCTL_QTSDONGLE_REQUEST_PAGE() As Int32

            Return CTL_CODE( _
                 FILE_DEVICE_UNKNOWN, _
                 QTSDONGLE_IOCTL_INDEX + &H16, _
                 METHOD_BUFFERED, _
                 FILE_ANY_ACCESS)
        End Function

        Private Shared Function _
             IOCTL_QTSDONGLE_GET_PAGE_FLAGS() As Int32

            Return CTL_CODE( _
                 FILE_DEVICE_UNKNOWN, _
                 QTSDONGLE_IOCTL_INDEX + &H17, _
                 METHOD_BUFFERED, _
                 FILE_ANY_ACCESS)
        End Function

        Private Shared Function _
             IOCTL_QTSDONGLE_GET_TWC_FLAG() As Int32

            Return CTL_CODE( _
                 FILE_DEVICE_UNKNOWN, _
                 QTSDONGLE_IOCTL_INDEX + &H18, _
                 METHOD_BUFFERED, _
                 FILE_ANY_ACCESS)
        End Function

        Private Shared Function _
             IOCTL_QTSDONGLE_GET_BUSY_FLAG() As Int32

            Return CTL_CODE( _
                 FILE_DEVICE_UNKNOWN, _
                 QTSDONGLE_IOCTL_INDEX + &H19, _
                 METHOD_BUFFERED, _
                 FILE_ANY_ACCESS)
        End Function

        Private Shared Function _
             IOCTL_QTSDONGLE_GET_EEPROM_ERROR_FLAG() As Int32

            Return CTL_CODE( _
                 FILE_DEVICE_UNKNOWN, _
                 QTSDONGLE_IOCTL_INDEX + &H1A, _
                 METHOD_BUFFERED, _
                 FILE_ANY_ACCESS)
        End Function

        Private Shared Function _
             IOCTL_QTSDONGLE_SOFT_RESET() As Int32

            Return CTL_CODE( _
                 FILE_DEVICE_UNKNOWN, _
                 QTSDONGLE_IOCTL_INDEX + &H1C, _
                 METHOD_BUFFERED, _
                 FILE_ANY_ACCESS)
        End Function

        Private Shared Function _
             IOCTL_QTSDONGLE_SEND_PORT() As Int32

            Return CTL_CODE( _
                 FILE_DEVICE_UNKNOWN, _
                 QTSDONGLE_IOCTL_INDEX + &H1D, _
                 METHOD_BUFFERED, _
                 FILE_ANY_ACCESS)
        End Function

        Private Shared Function _
             IOCTL_QTSDONGLE_GET_FLAGS() As Int32

            Return CTL_CODE( _
                 FILE_DEVICE_UNKNOWN, _
                 QTSDONGLE_IOCTL_INDEX + &H1E, _
                 METHOD_BUFFERED, _
                 FILE_ANY_ACCESS)
        End Function

        Private Shared Function _
             IOCTL_QTSDONGLE_SET_MODE() As Int32

            Return CTL_CODE( _
                 FILE_DEVICE_UNKNOWN, _
                 QTSDONGLE_IOCTL_INDEX + &H1F, _
                 METHOD_BUFFERED, _
                 FILE_ANY_ACCESS)
        End Function

        Private Shared Function _
             IOCTL_QTSDONGLE_GET_MODE() As Int32

            Return CTL_CODE( _
                 FILE_DEVICE_UNKNOWN, _
                 QTSDONGLE_IOCTL_INDEX + &H20, _
                 METHOD_BUFFERED, _
                 FILE_ANY_ACCESS)
        End Function

        Private Shared Function _
             IOCTL_QTSDONGLE_SET_STRICT() As Int32

            Return CTL_CODE( _
                 FILE_DEVICE_UNKNOWN, _
                 QTSDONGLE_IOCTL_INDEX + &H22, _
                 METHOD_BUFFERED, _
                 FILE_ANY_ACCESS)
        End Function

    End Class

    Private Function OpenUsbDev(ByRef dongle As Device) As SmartDnglError

        ' WINNT.h
        Const FILE_SHARE_READ As Integer = &H1
        Const FILE_SHARE_WRITE As Integer = &H2

        ' WinBase.h
        Const OPEN_EXISTING As Integer = 3

        Const DIGCF_INTERFACEDEVICE As Integer = &H10
        Const DIGCF_PRESENT As Integer = &H2

        Const ERROR_NO_MORE_ITEMS As Integer = 259
        Const ERROR_INSUFFICIENT_BUFFER As Integer = 122

        Const INVALID_HANDLE_VALUE As Integer = -1

        Dim status As Integer
        Dim success As Boolean
        Dim deviceInfoSet As Integer
        Dim devInterfaceData As Win32Methods.SP_DEVICE_INTERFACE_DATA
        Dim devInterfaceDetailData As New Win32Methods.SP_DEVICE_INTERFACE_DETAIL_DATA
        Dim devInterface As Integer
        Dim requiredLength As Integer
        Dim devicePath As String
        Dim tmp As IntPtr
        Dim buffer() As Byte


        Dim dnglError As SmartDnglError = SmartDnglError.UskOk

        devInterfaceData = New Win32Methods.SP_DEVICE_INTERFACE_DATA
        devInterfaceData.cbSize = _
            Marshal.SizeOf(GetType(Win32Methods.SP_DEVICE_INTERFACE_DATA))

        deviceInfoSet = Win32Methods.SetupDiGetClassDevs( _
            ClassGuid, _
            IntPtr.Zero, _
            IntPtr.Zero, _
            DIGCF_PRESENT Or DIGCF_INTERFACEDEVICE)

        If deviceInfoSet = INVALID_HANDLE_VALUE Then
            Return SmartDnglError.UskErrOpen
        End If

        ' Get the FIRST interface, should be only one anyways
        devInterface = 0
        success = Win32Methods.SetupDiEnumDeviceInterfaces( _
            deviceInfoSet, _
            IntPtr.Zero, _
            ClassGuid, _
            devInterface, _
            devInterfaceData)

        If success Then

            ' Get the required buffer size, allocate pointer of returned
            ' size and call SetupDiGetDeviceInterfaceDetail() again.
            success = Win32Methods.SetupDiGetDeviceInterfaceDetail( _
                deviceInfoSet, _
                devInterfaceData, _
                IntPtr.Zero, _
                0, _
                requiredLength, _
                IntPtr.Zero)

            If Marshal.GetLastWin32Error() <> ERROR_INSUFFICIENT_BUFFER Then
                Win32Methods.SetupDiDestroyDeviceInfoList(deviceInfoSet)
                Return SmartDnglError.UskErrOpen
            End If

            devInterfaceDetailData.cbSize = _
                Marshal.SizeOf(GetType(Win32Methods.SP_DEVICE_INTERFACE_DETAIL_DATA))

            tmp = Marshal.AllocHGlobal(requiredLength)
            buffer = New Byte(requiredLength - 1) {}

            Marshal.StructureToPtr(devInterfaceDetailData, tmp, False)
            Marshal.Copy(tmp, buffer, 0, requiredLength)
            Marshal.FreeHGlobal(tmp)

            success = Win32Methods.SetupDiGetDeviceInterfaceDetail( _
                deviceInfoSet, _
                devInterfaceData, _
                buffer, _
                requiredLength, _
                requiredLength, _
                IntPtr.Zero)


            If success Then

                ' Retrieve the DevicePath from
                ' SP_DEVICE_INTERFACE_DETAIL_DATA
                ' 4 == Offset of first DevicePath character
                Dim enc As New System.Text.UnicodeEncoding

                devicePath = enc.GetString(buffer, 4, requiredLength - 4)

                ' Open handle to physical device
                dongle.device = Win32Methods.CreateFile( _
                    devicePath.ToCharArray(), _
                    0, _
                    FILE_SHARE_READ Or FILE_SHARE_WRITE, _
                    IntPtr.Zero, _
                    OPEN_EXISTING, _
                    0, _
                    IntPtr.Zero)

                If Marshal.GetLastWin32Error() <> 0 Then
                    dnglError = SmartDnglError.UskErrOpen
                End If
            Else
                dnglError = SmartDnglError.UskErrOpen
            End If
        Else
            status = Marshal.GetLastWin32Error()
            If status = ERROR_NO_MORE_ITEMS Then
                dnglError = SmartDnglError.UskErrNoDevice
            Else
                dnglError = SmartDnglError.UskErrOpen
            End If
        End If

        success = Win32Methods.SetupDiDestroyDeviceInfoList(deviceInfoSet)

        Return dnglError
    End Function

    ' a = a + b + carry_in   function returns true if carry out generated
    Private Shared Function add_bytes(ByRef a As Byte, ByVal b As Byte, ByVal carry_in As Boolean) As Boolean

        Dim x As Int16, y As Int16, z As Int16
        Dim tmp(1) As Byte


        tmp(0) = a
        y = BitConverter.ToInt16(tmp, 0)

        tmp(0) = b
        z = BitConverter.ToInt16(tmp, 0)

        x = y + z

        If carry_in Then
            x = CShort(x + 1)
        End If

        If (x And &H100) > 0 Then
            x = CShort(x And &HFF)
            add_bytes = True
        Else
            add_bytes = False
        End If

        tmp = BitConverter.GetBytes(x)

        a = tmp(0)
    End Function

    Private Shared Function shift_byte(ByRef a As Byte, ByVal carry_in As Boolean) As Boolean

        Dim x As Integer
        Dim tmp(4) As Byte


        tmp(0) = a
        x = BitConverter.ToInt32(tmp, 0)

        ' Return true if carry out
        If (x And &H1) > 0 Then
            x = x And Not &H1
            shift_byte = True
        Else
            shift_byte = False
        End If

        x = CInt(x / 2) ' shift right

        ' Carry in
        If carry_in Then
            x += &H80
        End If

        tmp = BitConverter.GetBytes(x)

        a = tmp(0)
    End Function

    ' Shift 128 bits, (var64_2, var64_3)
    ' Returns true if a '1' was shifted out.
    Private Shared Function logical_right_shift_product( _
        ByRef var64_2 As Byte(), _
        ByRef var64_3 As Byte() _
        ) As Boolean

        Dim i As Integer
        Dim carry As Boolean


        carry = False

        For i = 0 To 7
            carry = shift_byte(var64_2(i), carry)
        Next i

        For i = 0 To 7
            carry = shift_byte(var64_3(i), carry)
        Next i

        Return carry
    End Function

    ' Add var64_1 and var64_2, result in var64_2, ignore carry out
    Private Shared Sub add_64X64( _
        ByRef var64_1 As Byte(), _
        ByRef var64_2 As Byte())

        Dim i As Integer
        Dim carry As Boolean

        carry = False

        For i = 7 To 0 Step -1
            carry = add_bytes(var64_2(i), var64_1(i), carry)
        Next i

    End Sub

    Private Shared Function Multiply(ByVal Multiplier As Int64, ByVal Multiplicand As Int64) As Int64

        Dim i As Integer, j As Integer
        Dim tmp(7) As Byte

        ' 64 bit math
        Dim var64_1(7) As Byte
        Dim var64_2(7) As Byte
        Dim var64_3(7) As Byte


        ' Convert big endian to little endian
        tmp = BitConverter.GetBytes(Multiplicand)
        j = 7
        For i = 0 To 7
            var64_1(j) = tmp(i)
            j -= 1
        Next i

        For i = 0 To 7
            var64_2(1) = &H0
        Next i

        ' Convert big endian to little endian
        tmp = BitConverter.GetBytes(Multiplier)
        j = 7
        For i = 0 To 7
            var64_3(j) = tmp(i)
            j -= 1
        Next i

        ' Shift and test multiplication
        For i = 0 To 63
            If logical_right_shift_product(var64_2, var64_3) Then
                add_64X64(var64_1, var64_2)
            End If
        Next

        ' Final shift
        logical_right_shift_product(var64_2, var64_3)

        ' Convert back to big endian
        j = 7
        For i = 0 To 7
            tmp(i) = var64_3(j)
            j -= 1
        Next i

        Return BitConverter.ToInt64(tmp, 0)
    End Function

    Private Shared Function Add(ByVal Addend1 As Int64, ByVal Addend2 As Int64) As Int64

        Dim i As Integer, j As Integer
        Dim tmp(7) As Byte

        ' 64 bit math
        Dim var64_1(7) As Byte
        Dim var64_2(7) As Byte
        Dim var64_3(7) As Byte


        ' Convert big endian to little endian
        tmp = BitConverter.GetBytes(Addend1)
        j = 7
        For i = 0 To 7
            var64_1(j) = tmp(i)
            j -= 1
        Next i

        ' Convert big endian to little endian
        tmp = BitConverter.GetBytes(Addend2)
        j = 7
        For i = 0 To 7
            var64_2(j) = tmp(i)
            j -= 1
        Next i

        add_64X64(var64_1, var64_2)

        ' Convert back to big endian
        j = 7
        For i = 0 To 7
            tmp(i) = var64_2(j)
            j -= 1
        Next i

        Return BitConverter.ToInt64(tmp, 0)
    End Function

    Public Enum SmartDnglError
        UskOk = 0
        UskErrOpen
        UskErrIsOpen
        UskErrMem
        UskErrEepromSize
        UskErr5
        UskErrSecurityStateTimeout
        UskErrKeyStateTimeout
        UskErrTwcResetTimeout
        UskErrBusyTimeout
        UskErrNotEnabled
        UskErrEepromReadTimeout
        UskErrEepromRead
        UskErrEepromWrite
        UskErrUskValidate
        UskErrHostValidate
        UskErrDisable
        UskErrDataFormat
        UskErrIoctlGetPipeInfo
        UskErrIoctlGetDeviceDescriptor
        UskErrIoctlGetConfigDescriptor
        UskErrIoctlRegisterNotifyEvent
        UskErrIoctlResetDevice
        UskErrIoctlResetPipe
        UskErrIoctlGetDriverRevision
        UskErrIoctlGetStringDescriptor
        UskErr26
        UskErrIoctlSendKey
        UskErrIoctlGetKey
        UskErrIoctlWritePort
        UskErrIoctlReadPort
        UskErrIoctlReadEeprom
        UskErrIoctlWriteEeprom
        UskErrIoctlGetSecurityState
        UskErrIoctlResetSecurityState
        UskErrIoctlGetKeyState
        UskErrIoctlResetKeyState
        UskErrIoctlSendHostVerifyRequest
        UskErrIoctlRequestPage
        UskErrIoctlGetPageFlags
        UskErrIoctlGetTwcFlag
        UskErrIoctlGetBusyFlag
        UskErrIoctlGetEepromErrorFlag
        UskErr43
        UskErr44
        UskErrIoctlQtsDongleSendPort
        UskErrIoctlQtsDongleGetFlags
        UskErrIoctlSetMode
        UskErrIoctlGetMode
        UskErrNoDevice

        UskErrBadArgument = 63

        UskErr80 = 80
        UskErrIoctlSetStrict
        UskErrRequiresNewerDriver
        UskErrRequiresNewerSmartDongleVersion
        UskWarkCryptPageBoundary
        UskWarnCryptPageBoundary

        UskResetComeready

        UskErrUserDataVerify

        UskErrUnknown = 255
    End Enum

    Public Shared Function GetErrorString( _
        ByVal dnglError As SmartDnglError) As String

        Select Case dnglError
            Case SmartDnglError.UskOk
                Return "no errors"

            Case SmartDnglError.UskErrOpen
                Return "Error: Can't open SmartDongle."

            Case SmartDnglError.UskErrIsOpen
                Return "Error: SmartDongle is already open."

                ' Memory allocation error is a system error
            Case SmartDnglError.UskErrMem
                Return "Error: Memory allocation error."

            Case SmartDnglError.UskErrEepromSize
                Return "Error: Address offset or buffer exceeds SmartDongle memory size."

            Case SmartDnglError.UskErr5
                Return "Error deprecated or reserved, this is an incorrect error value"

            Case SmartDnglError.UskErrSecurityStateTimeout
                Return "Error: Timeout waiting for security state change."

            Case SmartDnglError.UskErrKeyStateTimeout
                Return "Error: Timeout waiting for key state change"

            Case SmartDnglError.UskErrTwcResetTimeout
                Return "Error: Timeout waiting for TWC to reset"

            Case SmartDnglError.UskErrBusyTimeout
                Return "Error: Timeout waiting for job in firmware task loop to finish."

            Case SmartDnglError.UskErrNotEnabled
                Return "Error: Illegal command, SmartDongle needs to be unlocked first."

            Case SmartDnglError.UskErrEepromReadTimeout
                Return "Error: Timeout waiting for SmartDongle memory read."

            Case SmartDnglError.UskErrEepromRead
                Return "Error: SmartDongle memory Read error."

            Case SmartDnglError.UskErrEepromWrite
                Return "Error: SmartDongle memory Write error."

            Case SmartDnglError.UskErrUskValidate
                Return "Error: SmartDongle failed host challenge"

            Case SmartDnglError.UskErrHostValidate
                Return "Error: Bad keys, could not unlock SmartDongle"

            Case SmartDnglError.UskErrDisable
                Return "Error: Unable to lock SmartDongle."

            Case SmartDnglError.UskErrDataFormat
                Return "Error: Format of Serial Number read from SmartDongle is bad"

                '
                ' IOCTL SmartDnglErrors
                '
            Case SmartDnglError.UskErrIoctlGetPipeInfo
                Return "Error: IOCTL failed: IOCTL_GET_PIPE_INFO"

            Case SmartDnglError.UskErrIoctlGetDeviceDescriptor
                Return "Error: IOCTL failed: IOCTL_GET_DEVICE_DESCRIPTOR"

            Case SmartDnglError.UskErrIoctlGetConfigDescriptor
                Return "Error: IOCTL failed: IOCTL_GET_CONFIG_DESCRIPTOR"

            Case SmartDnglError.UskErrIoctlRegisterNotifyEvent
                Return "Error: IOCTL failed: IOCTL_REGISTER_NOTIFY_EVENT"

            Case SmartDnglError.UskErrIoctlResetDevice
                Return "Error: IOCTL failed: IOCTL_RESET_DEVICE"

            Case SmartDnglError.UskErrIoctlResetPipe
                Return "Error: IOCTL failed: IOCTL_RESET_PIPE"

            Case SmartDnglError.UskErrIoctlGetDriverRevision
                Return "Error: IOCTL failed: IOCTL_GET_DRIVER_REVISION"

            Case SmartDnglError.UskErrIoctlGetStringDescriptor
                Return "Error: IOCTL failed: IOCTL_GET_STRING_DESCRIPTOR"

            Case SmartDnglError.UskErr26
                Return "Error deprecated or reserved, this is an incorrect error value"

            Case SmartDnglError.UskErrIoctlSendKey
                Return "Error: IOCTL failed: IOCTL_SEND_KEY"

            Case SmartDnglError.UskErrIoctlGetKey
                Return "Error: IOCTL failed: IOCTL_GET_KEY"

            Case SmartDnglError.UskErrIoctlWritePort
                Return "Error: IOCTL failed: IOCTL_WRITE_PORT"

            Case SmartDnglError.UskErrIoctlReadPort
                Return "Error: IOCTL failed: IOCTL_READ_PORT"

            Case SmartDnglError.UskErrIoctlReadEeprom
                Return "Error: IOCTL failed: IOCTL_READ_EEPROM"

            Case SmartDnglError.UskErrIoctlWriteEeprom
                Return "Error: IOCTL failed: IOCTL_WRITE_EEPROM"

            Case SmartDnglError.UskErrIoctlGetSecurityState
                Return "Error: IOCTL failed: IOCTL_GET_SECURITY_STATE"

            Case SmartDnglError.UskErrIoctlResetSecurityState
                Return "Error: IOCTL failed: IOCTL_RESET_SECURITY_STATE"

            Case SmartDnglError.UskErrIoctlGetKeyState
                Return "Error: IOCTL failed: IOCTL_GET_KEY_STATE"

            Case SmartDnglError.UskErrIoctlResetKeyState
                Return "Error: IOCTL failed: IOCTL_RESET_KEY_STATE"

            Case SmartDnglError.UskErrIoctlSendHostVerifyRequest
                Return "Error: IOCTL failed: IOCTL_SEND_HOST_VERIFY_REQUEST"

            Case SmartDnglError.UskErrIoctlRequestPage
                Return "Error: IOCTL failed: IOCTL_REQUEST_PAGE"

            Case SmartDnglError.UskErrIoctlGetPageFlags
                Return "Error: IOCTL failed: IOCTL_GET_PAGE_FLAGS"

            Case SmartDnglError.UskErrIoctlGetTwcFlag
                Return "Error: IOCTL failed: IOCTL_GET_TWC_FLAG"

            Case SmartDnglError.UskErrIoctlGetBusyFlag
                Return "Error: IOCTL failed: IOCTL_GET_BUSY_FLAG"

            Case SmartDnglError.UskErrIoctlGetEepromErrorFlag
                Return "Error: IOCTL failed: IOCTL_GET_EEPROM_ERROR_FLAG"

            Case SmartDnglError.UskErr43
                Return "Error deprecated or reserved, this is an incorrect error value"

            Case SmartDnglError.UskErr44
                Return "Error deprecated or reserved, this is an incorrect error value"

            Case SmartDnglError.UskErrIoctlQtsDongleSendPort
                Return "Error: IOCTL failed: IOCTL_QTSDONGLE_SEND_PORT"

            Case SmartDnglError.UskErrIoctlQtsDongleGetFlags
                Return "Error: IOCTL failed: IOCTL_QTSDONGLE_GET_FLAGS"

            Case SmartDnglError.UskErrIoctlSetMode
                Return "Error: IOCTL failed: IOCTL_SET_MODE"

            Case SmartDnglError.UskErrIoctlGetMode
                Return "Error: IOCTL failed: IOCTL_GET_MODE"

                ' No device found.
            Case SmartDnglError.UskErrNoDevice
                Return "Error: No SmartDongles found."

            Case SmartDnglError.UskErrBadArgument
                Return "Error: One or more function arguments incorrect."

            Case SmartDnglError.UskErr80
                Return "Error deprecated or reserved, this is an incorrect error value"


                ' For SmartDongle v1.50
            Case SmartDnglError.UskErrIoctlSetStrict
                Return "Error: IOCTL failed: IOCTL_QTSDONGLE_SET_STRICT"

            Case SmartDnglError.UskErrRequiresNewerDriver
                Return "Error: An attempt to use 2.3 driver functionality with an older driver"

            Case SmartDnglError.UskErrRequiresNewerSmartDongleVersion
                Return "Error: An attempt to use new 1.50 SmartDongle functionality with an older dongle."

            Case SmartDnglError.UskWarkCryptPageBoundary
                Return "Warning: Data written to user memory after encryption mode change was not on a page boundary."

            Case SmartDnglError.UskWarnCryptPageBoundary
                Return "Warning: Encrypted data written to user memory was not on a page boundary."

            Case SmartDnglError.UskResetComeready
                Return "Warning: SmartDongle did not come ready after a reset."

            Case SmartDnglError.UskErrUserDataVerify
                Return "Error: Failed attempt to find SmartDongle with given keys and matching user data."

                ' Unknown error occurred.
            Case SmartDnglError.UskErrUnknown
                Return "An unknown error occurred"

            Case Else
                Return "An unknown error occurred"
        End Select
    End Function

End Class
