'    SmartDongle ClientForm.vb
'
'    MicroWorks, Inc.
'    2808 North Cole Road
'    Boise, ID 83704
'
'    www.smartdongle.com
'
'    ***** BEGIN LICENSE BLOCK *****
'
'    Version: MPL 1.1/GPL 2.0/LGPL 2.1
'
'    The contents of this file are subject to the Mozilla Public License
'    Version 1.1 (the "License"); you may not use this file except in
'    compliance with the License. You may obtain a copy of the License at
'    http://www.mozilla.org/MPL/
'
'    Software distributed under the License is distributed on an "AS IS" basis,
'    WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
'    for the specific language governing rights and limitations under the
'    License.
'
'    The Original Code is MicroWorks, Inc. code.
'
'    The Initial Developer of the Original Code is MicroWorks, Inc. Portions
'    created by MicroWorks, Inc. are Copyright (C) 2010
'    MicroWorks, Inc. All Rights Reserved.
'
'    Contributor(s):
'
'    Alternatively, the contents of this file may be used under the terms of
'    either the GNU General Public License Version 2 or later (the "GPL"), or
'    the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
'    in which case the provisions of the GPL or the LGPL are applicable instead
'    of those above. If you wish to allow use of your version of this file only
'    under the terms of either the GPL or the LGPL, and not to allow others to
'    use your version of this file under the terms of the MPL, indicate your
'    decision by deleting the provisions above and replace them with the notice
'    and other provisions required by the GPL or the LGPL. If you do not delete
'    the provisions above, a recipient may use your version of this file under
'    the terms of any one of the MPL, the GPL or the LGPL.
'
'    ***** END LICENSE BLOCK *****

Imports System.Net.Sockets
Imports System.Text
Imports System.Net

Class TCPSrv

    Shared Sub Main()
        ' Server IP address separated by commas instead of periods.
        Dim myAddress() As Byte = {10, 0, 0, 101}
        ' Port to listen on. Client must connect to this port.
        Const portNumber As Integer = 8000
        ' Address on SmartDonagle where number of seats is stored.
        Dim seatsAddress As UInt32 = 11722
        ' Address on SmartDonagle where seat counter is stored.
        Dim counterAddress As UInt32 = 2808
        ' Your keys here
        Dim P1 As Int64, P2 As Int64
        P1 = &HEC6CC589AEFD1E75
        P2 = &HFCEC0A6A82747B3F

        Dim maxBytes As Integer = 1024

        ' Get number of available seats from SmartDongle
        Dim maxConnections As Integer = getNumSeats(P1, P2, seatsAddress)
        If Not (maxConnections > 0) Then
            Console.WriteLine("No available seats found.")
            Console.WriteLine("Connect valid SmartDongle with available seats.")
            Console.WriteLine("Press any key to exit.")
            Console.ReadKey()
            Exit Sub
        End If
        Dim currConnections As Integer = 0
        Dim ExitFlag As Boolean = False
        Dim serverIPAddress As IPAddress = New IPAddress(myAddress)
        Dim tcpListener As New TcpListener(serverIPAddress, portNumber)

        ' Begin listening for incoming connections
        tcpListener.Start()
        While (Not ExitFlag)
            Console.WriteLine("Waiting for connection...")
            Try
                ' Message received by client.
                Dim clientdata As StringBuilder = New StringBuilder()
                'Accept the pending client connection and return a TcpClient initialized for communication. 
                Dim tcpClient As TcpClient = tcpListener.AcceptTcpClient()
                ' Get the stream
                Dim networkStream As NetworkStream = tcpClient.GetStream()

                ' Read the clien message from the stream into a byte array.
                Dim recvBytes(maxBytes) As Byte
                Do
                    networkStream.Read(recvBytes, 0, 1)
                    clientdata.AppendFormat("{0}", Encoding.ASCII.GetString(recvBytes, 0, 1))
                Loop While networkStream.DataAvailable

                ' Determine server action based on client message.
                If StrComp(clientdata.ToString, "get seat") = 0 Then
                    If currConnections < maxConnections Then
                        Dim dnglError As Integer = writeCountToDongle(P1, P2, counterAddress)
                        If dnglError = 0 Then
                            currConnections += 1
                            Console.WriteLine("Connection accepted.")
                            Console.WriteLine((maxConnections - currConnections).ToString + " seats available.")
                            Dim responseString As String = "accept"
                            Dim sendBytes As [Byte]() = Encoding.ASCII.GetBytes(responseString)
                            networkStream.Write(sendBytes, 0, sendBytes.Length)
                            tcpClient.Close()
                        Else
                            Console.WriteLine("Connection rejected.")
                            Dim responseString As String = SmartDongle.GetErrorString(dnglError)
                            Console.WriteLine(responseString)
                            Dim sendBytes As [Byte]() = Encoding.ASCII.GetBytes(responseString)
                            networkStream.Write(sendBytes, 0, sendBytes.Length)
                            tcpClient.Close()
                        End If
                    Else
                        Console.WriteLine("Connection rejected. No available seats.")
                        Dim responseString As String = "reject"
                        Dim sendBytes As [Byte]() = Encoding.ASCII.GetBytes(responseString)
                        networkStream.Write(sendBytes, 0, sendBytes.Length)
                        tcpClient.Close()
                    End If
                ElseIf StrComp(clientdata.ToString, "return seat") = 0 Then
                    If currConnections > 0 Then
                        currConnections -= 1
                    End If
                    Console.WriteLine("Seat returned.")
                    Console.WriteLine((maxConnections - currConnections).ToString + " seats available.")
                    Dim responseString As String = "returned"
                    Dim sendBytes As [Byte]() = Encoding.ASCII.GetBytes(responseString)
                    networkStream.Write(sendBytes, 0, sendBytes.Length)
                    tcpClient.Close()
                Else
                    Console.WriteLine("Connection rejected.")
                    Console.WriteLine("Unknown message received from client")
                    Dim responseString As String = "unknown"
                    Dim sendBytes As [Byte]() = Encoding.ASCII.GetBytes(responseString)
                    networkStream.Write(sendBytes, 0, sendBytes.Length)
                    tcpClient.Close()
                End If
            Catch e As Exception
                Console.WriteLine(e.ToString())
                Console.ReadLine()
            End Try
        End While
        tcpListener.Stop()
        ' Pause before exiting so user can read console.
        Console.WriteLine("Press any key to exit.")
        Console.ReadKey()
    End Sub

    ' Reads SmartDongle at address = seatsAddress
    ' to determine number of available seats.
    Shared Function getNumSeats(ByVal P1 As Int64, ByVal P2 As Int64, ByVal seatsAddress As UInt32) As Integer
        Dim buffSize As Integer = 1
        Dim numSeats As Integer = 0
        Dim dongle As New SmartDongle()
        Dim dnglError As SmartDongle.SmartDnglError
        Dim buff() As Byte = New Byte(buffSize) {}

        ' Check if valid SmartDongle is present.
        '
        dnglError = dongle.Read(P1, P2, seatsAddress, buff, buffSize)

        If dnglError <> SmartDongle.SmartDnglError.UskOk Then
            System.Console.WriteLine(SmartDongle.GetErrorString(dnglError))
        Else
            System.Console.WriteLine("Success: Valid SmartDongle found")
            numSeats = buff(0)
            Console.WriteLine(numSeats.ToString + " max seats available.")
        End If

        Return numSeats
    End Function

    ' Reads the current number of times a seat has been granted
    ' increments the count by 1 and rewrites the count to the
    ' SmartDongle at address = counterAddress.
    ' Also performs a check to ensure that a valid SmartDongle
    ' is still present on the server.
    Shared Function writeCountToDongle(ByVal P1 As Int64, ByVal P2 As Int64, ByVal counterAddress As UInt32) As Integer
        Dim count As Int32
        Dim buffSize As Integer = 4
        Dim dongle As New SmartDongle()
        Dim dnglError As SmartDongle.SmartDnglError
        Dim buff() As Byte = New Byte(buffSize) {}

        ' Check if valid SmartDongle is present.
        dnglError = dongle.Read(P1, P2, counterAddress, buff, buffSize)

        If dnglError <> SmartDongle.SmartDnglError.UskOk Then
            System.Console.WriteLine(SmartDongle.GetErrorString(dnglError))
        Else
            System.Console.WriteLine("Success: Valid SmartDongle found")
            count = BitConverter.ToInt32(buff, 0)
            count += 1
            Console.WriteLine("Current Count: " + count.ToString)
            buff = BitConverter.GetBytes(count)
            dongle.Write(P1, P2, counterAddress, buff, buffSize)
        End If
        Return dnglError
    End Function

End Class
